#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dash import dcc
import dash_bootstrap_components as dbc
from dash import html
from .app import app
import os
import sys
import base64

from dash import Input, Output, State

__all__ = ["upload", "UPLOAD_DIRECTORY"]

UPLOAD_DIRECTORY = os.path.join(os.getcwd(), "config_gui_temp")
sys.path.append(UPLOAD_DIRECTORY)

upload = html.Div(
    [
        html.P(
            [
                "Please upload below your files",
                " or ",
                dbc.Button(
                    "start from empty config",
                    color="primary",
                    className="me-1",
                    id="button-empty-config",
                ),
                html.P(id="config-initial-status"),
            ]
        ),
        dcc.Upload(
            id="upload-config",
            children=html.Div(
                [
                    "Python file that generates global variable ",
                    html.Strong("configuration "),
                    "(Drag and Drop or Click to select)",
                ]
            ),
            style={
                "width": "100%",
                "height": "60px",
                "lineHeight": "60px",
                "borderWidth": "1px",
                "borderStyle": "dashed",
                "borderRadius": "5px",
                "textAlign": "center",
                "margin": "10px",
            },
            # Allow multiple files to be uploaded
            multiple=False,
        ),
        html.Div(id="output-data-upload"),
        dcc.Upload(
            id="upload-config-edits",
            children=html.Div(
                [
                    "Optionally, any previously generated ",
                    html.Strong("config_edits.py"),
                    "(Drag and Drop or Click to select)",
                ]
            ),
            style={
                "width": "100%",
                "height": "60px",
                "lineHeight": "60px",
                "borderWidth": "1px",
                "borderStyle": "dashed",
                "borderRadius": "5px",
                "textAlign": "center",
                "margin": "10px",
            },
            # Allow multiple files to be uploaded
            multiple=False,
        ),
        html.Div(id="output-data-edits-upload"),
    ]
)


@app.callback(
    [Output("output-data-upload", "children")],
    Input("upload-config", "contents"),
    State("upload-config", "filename"),
    State("upload-config", "last_modified"),
)
def upload_init_config(contents, filename, last_modified):
    if contents is not None:
        init_edits_file()
        init_final_config_file()
        save_file(os.path.join(UPLOAD_DIRECTORY, "config_initial.py"), contents)
        print("OK")
        return [html.P(f"Uploaded {filename}")]
    return [""]


@app.callback(
    [Output("output-data-edits-upload", "children")],
    Input("upload-config-edits", "contents"),
    State("upload-config-edits", "filename"),
    State("upload-config-edits", "last_modified"),
)
def upload_config_edits(contents, filename, last_modified):
    if contents is not None:
        save_file(os.path.join(UPLOAD_DIRECTORY, "config_edits.py"), contents)
        return [html.P(f"Uploaded {filename}")]
    return [""]


def save_file(name, content):
    """Decode and store a file uploaded with Plotly Dash."""
    data = content.encode("utf8").split(b";base64,")[1]
    with open(os.path.join(UPLOAD_DIRECTORY, name), "wb") as fp:
        fp.write(base64.decodebytes(data))


@app.callback(
    Output("config-initial-status", "children"),
    Input("button-empty-config", "n_clicks"),
    prevent_initial_call=True,
)
def enter_empty_config(click):
    init_empty_initial_config_file()
    init_edits_file()
    init_final_config_file()

    return "Starting from empty configuration..."


def init_edits_file():
    with open(os.path.join(UPLOAD_DIRECTORY, "config_edits.py"), "w") as fp:
        fp.write(
            """
from qualang_tools.config import *
from qualang_tools.config.server.config_editor import config_editor
from config_initial import configuration

setup = ConfigBuilder()

"""
        )


def init_empty_initial_config_file():
    with open(os.path.join(UPLOAD_DIRECTORY, "config_initial.py"), "w") as fp:
        fp.write(
            """
configuration = {
    "version": 1,
    "controllers": {},
    "elements": {},
    "pulses": {},
    "waveforms": {},
    "digital_waveforms": {},
    "integration_weights": {},
    "oscillators": {},
    "mixers": {},
}
"""
        )


def init_final_config_file():
    with open(os.path.join(UPLOAD_DIRECTORY, "config_final.py"), "w") as fp:
        fp.write(
            """
from config_edits import configuration, setup

configuration = setup.build(configuration)
"""
        )



from qualang_tools.wirer.connectivity.channel_spec import ChannelSpec
from qualang_tools.wirer.connectivity.types import QubitsType
from qualang_tools.wirer.connectivity.wiring_spec import WiringFrequency, WiringIOType, WiringLineType
from qualang_tools.wirer.connectivity.connectivity_base import ConnectivityBase
from qualang_tools.wirer.instruments.instrument_channel import (
    InstrumentChannelLfFemOutput,
    InstrumentChannelLfFemDigitalOutput,
)


class ConnectivityNVCenters(ConnectivityBase):
    """
    Represents the high-level wiring configuration for an NV center-based QPU setup.

    This class defines and stores placeholders for quantum elements (e.g., qubits and lasers)
    and specifies the wiring requirements for each of their control and readout lines. It enables
    the configuration of line types (e.g., drive, laser, readout), their I/O roles, and associated
    frequency domains (RF or DC), as well as constraints for channel allocation.

    The API is designed to model single NV centers, along with pairwise coupling mechanisms.
    """

    def add_nv_center(self, qubits: QubitsType):
        """
        Adds specifications (placeholders) for nv centers.

        This method configures the wiring specifications (placeholders) for a set of nv centers,
        including the qubit drive and laser. No channels are allocated at this stage.

        Args:
            qubits (QubitsType): The qubits to configure for nv centers.
        """
        self.add_laser(qubits)
        self.add_spcm(qubits)
        self.add_qubit_drive(qubits)

    def add_laser(self, qubits: QubitsType, triggered: bool = True, constraints: ChannelSpec = None):
        """
        Adds a specification (placeholder) for a laser for the specified qubits.

        This method configures a laser specification (placeholder) that can handle output,
        typically for reading out and initializing the state of qubits. It also allows optional
        triggering and constraints on which channel configurations can be allocated.

        No channels are allocated at this stage.

        Args:
            qubits (QubitsType): The qubits to associate with the laser.
            triggered (bool, optional): Whether the laser is triggered. Defaults to True.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.

        Returns:
            A wiring specification (placeholder) for the laser.
        """
        elements = self._make_qubit_elements(qubits)

        # check constraints to assign correct wiring frequency
        if constraints and constraints.channel_templates:
            channel_types = {type(ch) for ch in constraints.channel_templates}
            if InstrumentChannelLfFemOutput in channel_types:
                wiring_freq = WiringFrequency.DC
            elif InstrumentChannelLfFemDigitalOutput in channel_types:
                wiring_freq = WiringFrequency.DO
            else:
                raise ValueError("Invalid channel constraint.")
        else:
            wiring_freq = WiringFrequency.DC

        return self.add_wiring_spec(
            wiring_freq,
            WiringIOType.OUTPUT,
            WiringLineType.LASER,
            triggered,
            constraints,
            elements,
        )

    def add_spcm(self, qubits: QubitsType, triggered: bool = False, constraints: ChannelSpec = None):
        """
        Adds a specification (placeholder) for a spcm for the specified qubits.

        This method configures a spcm specification (placeholder) that can handle input,
        typically for reading out the state of qubits. It also allows optional
        triggering and constraints on which channel configurations can be allocated.

        No channels are allocated at this stage.

        Args:
            qubits (QubitsType): The qubits to associate with the laser.
            triggered (bool, optional): Whether the readout is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.

        Returns:
            A wiring specification (placeholder) for the spcm.
        """
        elements = self._make_qubit_elements(qubits)
        return self.add_wiring_spec(
            WiringFrequency.DC,
            WiringIOType.INPUT,
            WiringLineType.SPCM,
            triggered,
            constraints,
            elements,
        )

    def add_qubit_drive(self, qubits: QubitsType, triggered: bool = False, constraints: ChannelSpec = None):
        """
        Adds specifications (placeholders) for drive lines for the specified qubits.

        This method configures the qubit drive line specifications (placeholders), which are typically used to apply
        control signals to qubits. It allows optional triggering and constraints on which channel configurations
        can be allocated for this line.

        No channels are allocated at this stage.


        Args:
            qubits (QubitsType): The qubits to configure the drive lines for.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.

        Returns:
            A wiring specification (placeholder) for the qubit drive lines.
        """
        elements = self._make_qubit_elements(qubits)
        return self.add_wiring_spec(
            WiringFrequency.RF,
            WiringIOType.OUTPUT,
            WiringLineType.DRIVE,
            triggered,
            constraints,
            elements,
        )



from dataclasses import dataclass
from typing import Union, Literal, Callable


@dataclass(eq=False)
class InstrumentChannel:
    con: int
    port: int
    slot: Union[None, int] = None

    def __str__(self):
        return f'({", ".join([f"con{self.con}", f"{self.slot}" if self.slot else "", str(self.port)])})'

    def make_channel_filter(self) -> Callable[["InstrumentChannel"], bool]:
        return lambda channel: (
            (self.con is None or self.con == channel.con)
            and (self.slot is None or self.slot == channel.slot)
            and (self.port is None or self.port == channel.port)
        )


@dataclass(eq=False)
class InstrumentChannelInput:
    io_type: Literal["input", "output"] = "input"


@dataclass(eq=False)
class InstrumentChannelOutput:
    io_type: Literal["input", "output"] = "output"


@dataclass(eq=False)
class InstrumentChannelDigital:
    signal_type = "digital"


@dataclass(eq=False)
class InstrumentChannelAnalog:
    signal_type = "analog"


InstrumentIdType = Literal["lf-fem", "mw-fem", "opx+", "octave", "external-mixer"]


@dataclass(eq=False)
class InstrumentChannelLfFem:
    instrument_id: InstrumentIdType = "lf-fem"


@dataclass(eq=False)
class InstrumentChannelMwFem:
    instrument_id: InstrumentIdType = "mw-fem"


@dataclass(eq=False)
class InstrumentChannelOpxPlus:
    instrument_id: InstrumentIdType = "opx+"


@dataclass(eq=False)
class InstrumentChannelOctave:
    instrument_id: InstrumentIdType = "octave"


@dataclass(eq=False)
class InstrumentChannelExternalMixer:
    instrument_id: InstrumentIdType = "external-mixer"


@dataclass(eq=False)
class InstrumentChannelLfFemInput(
    InstrumentChannelAnalog, InstrumentChannelLfFem, InstrumentChannelInput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelLfFemOutput(
    InstrumentChannelAnalog, InstrumentChannelLfFem, InstrumentChannelOutput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelLfFemDigitalOutput(
    InstrumentChannelDigital, InstrumentChannelLfFem, InstrumentChannelOutput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelMwFemInput(
    InstrumentChannelAnalog, InstrumentChannelMwFem, InstrumentChannelInput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelMwFemOutput(
    InstrumentChannelAnalog, InstrumentChannelMwFem, InstrumentChannelOutput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelMwFemDigitalOutput(
    InstrumentChannelDigital, InstrumentChannelMwFem, InstrumentChannelOutput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelOpxPlusInput(
    InstrumentChannelAnalog, InstrumentChannelOpxPlus, InstrumentChannelInput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelOpxPlusOutput(
    InstrumentChannelAnalog, InstrumentChannelOpxPlus, InstrumentChannelOutput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelOpxPlusDigitalOutput(
    InstrumentChannelDigital, InstrumentChannelOpxPlus, InstrumentChannelOutput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelExternalMixerInput(
    InstrumentChannelAnalog, InstrumentChannelExternalMixer, InstrumentChannelInput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelExternalMixerOutput(
    InstrumentChannelAnalog, InstrumentChannelExternalMixer, InstrumentChannelOutput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelOctaveInput(
    InstrumentChannelAnalog, InstrumentChannelOctave, InstrumentChannelInput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelOctaveOutput(
    InstrumentChannelAnalog, InstrumentChannelOctave, InstrumentChannelOutput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelOctaveDigitalInput(
    InstrumentChannelDigital, InstrumentChannelOctave, InstrumentChannelInput, InstrumentChannel
):
    pass


@dataclass(eq=False)
class InstrumentChannelExternalMixerDigitalInput(
    InstrumentChannelDigital, InstrumentChannelExternalMixer, InstrumentChannelInput, InstrumentChannel
):
    pass


AnyInstrumentChannel = Union[
    InstrumentChannelLfFemInput,
    InstrumentChannelLfFemOutput,
    InstrumentChannelMwFemInput,
    InstrumentChannelMwFemOutput,
    InstrumentChannelOpxPlusInput,
    InstrumentChannelOpxPlusOutput,
    InstrumentChannelOctaveInput,
    InstrumentChannelOctaveOutput,
    InstrumentChannelExternalMixerInput,
    InstrumentChannelExternalMixerOutput,
]



from dataclasses import dataclass
from typing import Union
from copy import deepcopy


@dataclass(eq=False, frozen=False)
class Pulser:
    controller: int
    slot: Union[None, int] = None
    allocated = False

    def __eq__(self, other):
        if not isinstance(other, Pulser):
            return NotImplemented
        return self.controller == other.controller and self.slot == other.slot

    def allocate(self) -> bool:
        if self.allocated:
            return False
        self.allocated = True
        return True


class InstrumentPulsers:
    """
    Collection of "stack" Pulser objects.
    Each Pulser entry contains the Controller and FEM Slot on which the thread is running. This wrapper
    can be interacted with as if it were the underlying stack dictionary.
    """

    def __init__(self):
        self.stack = []

    def add(self, pulser: Pulser):
        if not isinstance(pulser, Pulser):
            raise TypeError("Only instances of Pulser can be added to the stack.")
        self.stack.append(pulser)

    def remove(self, pulser: Pulser):
        if not isinstance(pulser, Pulser):
            raise TypeError("Only instances of Pulser can be removed from the stack.")
        try:
            self.stack.remove(pulser)
        except ValueError:
            from ..wirer.wirer_exceptions import NotEnoughPulsersException

            raise NotEnoughPulsersException(
                f"Pulser {pulser} does not exist in the stack. All the pulsers were used up."
            )

    def remove_by_slot(self, controller: int, slot: int):
        """
        Removes the first pulser that match the given controller and slot.
        """
        for pulser in self.stack:
            if pulser.controller == controller and pulser.slot == slot:
                self.stack.remove(pulser)
                return
        from ..wirer.wirer_exceptions import NotEnoughPulsersException

        raise NotEnoughPulsersException(f"Not enough pulsers available for controller {controller} and slot {slot}.")

    def filter_by_controller(self, controller: int):
        """
        Returns a list of pulsers that match the given controller.
        """
        filtered_pulsers = [pulser for pulser in self.stack if pulser.controller == controller]
        return filtered_pulsers

    def filter_by_slot(self, controller: int, slot: int):
        """
        Returns a list of pulsers that match the given controller.
        """
        filtered_pulsers = [pulser for pulser in self.stack if pulser.controller == controller and pulser.slot == slot]
        return filtered_pulsers

    def filter_unallocated_by_slot(self, controller: int, slot: int):
        """
        Returns a list of pulsers that are not allocated.
        """
        return [pulser for pulser in self.filter_by_slot(controller, slot) if not pulser.allocated]

    def __iter__(self):
        return iter(self.stack)

    def __getitem__(self, index):
        if isinstance(index, tuple) and len(index) == 2:
            controller, slot = index
            # find the first pulser that matches the controller and slot
            for pulser in self.stack:
                if pulser.controller == controller and pulser.slot == slot:
                    return pulser
        elif isinstance(index, int):
            return self.stack[index]

    def deepcopy(self):
        new_stack = InstrumentPulsers()
        new_stack.stack = [deepcopy(pulser) for pulser in self.stack]
        return new_stack

    def replenish_pulsers(self, controller: int, slot: int = None):
        """
        Replenishes the pulsers for a given controller and slot.
        This is useful when pulsers are allocated and need to be reset.
        """
        if controller is not None and slot is not None:
            for pulser in self.filter_by_slot(controller, slot):
                pulser.allocated = False
        elif controller is not None and slot is None:
            for pulser in self.filter_by_slot(controller, slot):
                pulser.allocated = False
        elif controller is None and slot is None:
            for pulser in self.stack:
                pulser.allocated = False
        else:
            from ..wirer.wirer_exceptions import NotEnoughPulsersException

            raise NotEnoughPulsersException("Must specify either controller, both controller and slot or None.")



import inspect
import logging
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Rectangle
from qm.octave.octave_manager import _add_calibration_entries_to_config
from qm.octave.octave_mixer_calibration import MixerCalibrationResults
from qualang_tools.units import unit

logger = logging.getLogger(__name__)


class CalibrationResultPlotter:
    colors = ["black", "midnightblue", "navy", "darkblue", "mediumblue", "dodgerblue", "white", "lightgrey"]
    custom_cmap = LinearSegmentedColormap.from_list("custom_diverging", colors, N=256)
    u = unit(coerce_to_integer=True)

    def __init__(self, data: MixerCalibrationResults):
        """
        Initialize the CalibrationResultPlotter with calibration data.

        Args:
            data (MixerCalibrationResults): The mixer calibration results, output of qm.calibrate_element(element).
        """
        first_key = next(iter(data))
        self.lo_frequency = first_key[0]
        self.output_gain = first_key[1]
        self.if_data = data[(self.lo_frequency, self.output_gain)].image
        self.lo_data = data[(self.lo_frequency, self.output_gain)]
        self.if_frequency = next(iter(self.if_data))
        self.integration_length = self._get_integration_length()

    @staticmethod
    def _handle_zero_indices_and_masking(data):
        """
        Handle zero indices and masking for the given data.
        Replaces 0.0 by NaN and returns the indices of zero values.

        Args:
            data (np.ndarray): The data array to process.

        Returns:
            list: A list of tuples containing the indices of zero values.
        """
        zero_idxs = np.where((data == 0.0) | np.isnan(data))
        zero_list = list(zip(zero_idxs[0], zero_idxs[1]))

        mask = (data == 0.0) | np.isnan(data)
        data[mask] = np.nan

        return zero_list

    @staticmethod
    def _plot_scan(scan_x, scan_y, data_dbm, zero_list, width, height, xlabel, ylabel):
        """
        Plot the scan data.

        Args:
            scan_x (np.ndarray): The x-axis scan data.
            scan_y (np.ndarray): The y-axis scan data.
            data_dbm (np.ndarray): The data in dBm.
            zero_list (list): List of zero value indices.
            width (float): Width of the rectangles to plot.
            height (float): Height of the rectangles to plot.
            xlabel (str): Label for the x-axis.
            ylabel (str): Label for the y-axis.
        """
        plt.pcolor(scan_x, scan_y, data_dbm, cmap=CalibrationResultPlotter.custom_cmap)
        mesh = plt.pcolor(scan_x, scan_y, data_dbm, cmap=CalibrationResultPlotter.custom_cmap)
        mesh.set_array(data_dbm)
        plt.colorbar(mesh, label="Power [dBm]")
        ax = plt.gca()
        for ii, _ in enumerate(zero_list):
            rect = Rectangle(
                (scan_x[0][zero_list[ii][1]] - width / 2, scan_y[zero_list[ii][0]][0] - height / 2),
                width,
                height,
                facecolor="black",
                edgecolor="none",
            )
            ax.add_patch(rect)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.axis("equal")

    @staticmethod
    def _convert_to_dbm(volts):
        """
        Convert voltage to dBm assuming a 50 Ohm impedance.

        Args:
            volts (np.ndarray): The voltage values.

        Returns:
            np.ndarray: The converted dBm values.
        """
        return 10 * np.log10(volts / (50 * 2) * 1000)

    @staticmethod
    def _plot_fit_elliptical_levels(d, x0, y0, q_scan, i_scan):
        """
        Plot the fit elliptical levels.

        Args:
            d: The data containing fit information.
            x0: The x-coordinate of the fit minimum.
            y0: The y-coordinate of the fit minimum.
            q_scan: The 'Q' scan data.
            i_scan: The 'I' scan data.
        """
        r = (
            np.min(
                [
                    y0 - np.min(i_scan),
                    np.max(i_scan) - y0,
                    x0 - np.min(q_scan),
                    np.max(q_scan) - x0,
                ]
            )
            * 0.9
        )
        ais = np.linspace(0, 1, 361) * 2 * np.pi
        P = np.array([np.cos(ais), np.sin(ais)]) * r

        O = np.zeros_like(P)
        O[0] += x0
        O[1] += y0

        plt.plot(*(P + O), "k:", linewidth=0.5)

        r1, r2 = d.fit.pol_[3] ** -0.5, d.fit.pol_[5] ** -0.5
        r1, r2 = r1 / np.sqrt(r1 * r2), r2 / np.sqrt(r1 * r2)

        tc = np.cos(d.fit.theta)
        ts = np.sin(d.fit.theta)

        P = np.array([[r1, 0], [0, r2]]) @ P
        P = np.array([[tc, ts], [-ts, tc]]) @ P

        plt.plot(*(P + O), "k--", linewidth=0.5)

    def show_lo_leakage_calibration_result(self) -> matplotlib.figure.Figure:
        """
        Plots the results of the LO leakage calibration process.

        This method generates a series of subplots that visualize the coarse and fine scans of the LO calibration and
        returns the achieved LO suppression in dB when comparing the power without calibrating the mixers vs
        after calibration. The powers shown correspond to the power seen by the OPX+ inputs when the signal is routed
        internally through the Octave.

        Note: The power at the Octave RF output may vary up to 10dBm.

        Subplots:
        - Top Left: Summary of the current result and achieved LO suppression.
        - Top Right: Coarse scan of Q_dc vs I_dc with LO power in dBm.
        - Bottom Left: Fit error in dBm.
        - Bottom Right: Fine scan of Q_dc vs I_dc with LO power in dBm.

        Returns:
            matplotlib.figure.Figure: The generated figure.
        """

        fig = plt.figure(figsize=(11, 8.9))

        plt.subplot(222)

        d = self.lo_data.debug.coarse[0]  # Get the coarse scan data

        q_scan = d.q_scan * 1000  # convert to mV
        i_scan = d.i_scan * 1000
        zero_list = self._handle_zero_indices_and_masking(
            d.lo
        )  # Get 0.0 indices and mask them with NaN values before converting to dBm

        lo = self.u.demod2volts(d.lo, self.integration_length)
        lo_dbm = self._convert_to_dbm(lo)

        width = q_scan[0][1] - q_scan[0][0]  # Calculate the width and height of the rectangles to plot
        height = i_scan[1][0] - i_scan[0][0]

        self._plot_scan(q_scan, i_scan, lo_dbm, zero_list, width, height, "Q_dc (mV)", "I_dc (mV)")

        plt.text(
            np.min(q_scan) + 0.5 * width,
            np.max(i_scan) - 0.5 * height,
            f"coarse scan\nLO = {self.lo_frequency / 1e9:.3f}GHz",
            color="k",
            bbox=dict(facecolor="w", alpha=0.8),
            verticalalignment="top",
        )

        if d.fit is None:
            raise RuntimeError("Fitting failed!")

        x0, y0 = d.fit.x_min * 1000, d.fit.y_min * 1000  # Convert to mV

        self._plot_fit_elliptical_levels(d, x0, y0, q_scan, i_scan)

        plt.plot(x0, y0, "yo", markersize=8)
        plt.plot(x0, y0, "ro", markersize=4)

        plt.text(
            x0,
            y0 - height * 2,
            f"Q_dc={x0:.2f}mV\nI_dc={y0:.2f}mV",
            color="y",
            horizontalalignment="center",
            verticalalignment="top",
        )

        fine_debug_data = self.lo_data.debug.fine

        x0_ref, y0_ref = x0, y0

        d = fine_debug_data[0]

        x0_fine = d.fit.x_min * 1000 + x0_ref
        y0_fine = d.fit.y_min * 1000 + y0_ref

        plt.plot(x0_fine, y0_fine, "wo", markersize=1)

        fine_scale = np.max(d.i_scan) * 1000

        plt.plot(
            x0_ref + np.array([1, 1, -1, -1, 1]) * fine_scale,
            y0_ref + np.array([1, -1, -1, 1, 1]) * fine_scale,
            "w-.",
            linewidth=0.5,
        )

        plt.subplot(224)

        fine_q_scan = d.q_scan * 1000 + x0_ref
        fine_i_scan = d.i_scan * 1000 + y0_ref

        zero_list = self._handle_zero_indices_and_masking(d.lo)

        lo = self.u.demod2volts(d.lo, self.integration_length)
        lo_dbm = self._convert_to_dbm(lo)

        width = fine_q_scan[0][1] - fine_q_scan[0][0]
        height = fine_i_scan[1][0] - fine_i_scan[0][0]

        self._plot_scan(fine_q_scan, fine_i_scan, lo_dbm, zero_list, width, height, "Q_dc (mV)", "I_dc (mV)")

        x0, y0 = d.fit.x_min * 1000 + x0_ref, d.fit.y_min * 1000 + y0_ref

        self._plot_fit_elliptical_levels(d, x0, y0, fine_q_scan, fine_i_scan)

        plt.plot(x0, y0, "yo", markersize=8)
        plt.plot(x0, y0, "ro", markersize=4)

        t = plt.text(
            np.min(fine_q_scan) + 0.5 * width,
            np.max(fine_i_scan) - 0.5 * height,
            f"fine scan\nLO = {self.lo_frequency / 1e9:.3f}GHz",
            color="k",
            verticalalignment="top",
        )
        t.set_bbox(dict(facecolor="w", alpha=0.8))

        plt.text(
            x0,
            y0 - height * 2,
            f"Q_dc={x0:.2f}mV\nI_dc={y0:.2f}mV",
            color="y",
            horizontalalignment="center",
            verticalalignment="top",
        )

        plt.subplot(223)

        X, Y = d.q_scan, d.i_scan
        p = d.fit.pol

        iq_error = self.u.demod2volts(
            p[0] + p[1] * X + p[2] * Y + p[3] * X**2 + p[4] * X * Y + p[5] * Y**2 - d.lo, self.integration_length
        )
        iq_error_dbm = self._convert_to_dbm(np.abs(iq_error))

        plt.pcolor(
            fine_q_scan,
            fine_i_scan,
            iq_error_dbm,
            cmap=CalibrationResultPlotter.custom_cmap,
        )
        plt.xlabel("Q (mV)")
        plt.ylabel("I (mV)")
        plt.axis("equal")

        t = plt.text(
            np.min(fine_q_scan) + 0.5 * width,
            np.max(fine_i_scan) - 1.5 * height,
            "fit error",
            color="k",
        )
        t.set_bbox(dict(facecolor="w", alpha=0.8))

        plt.colorbar(label="Power [dBm]")

        plt.subplot(221)

        plt.xlim(0, 1)
        plt.ylim(0, 1)

        content = [
            "Current result:",
            f"      I_dc = {y0:.02f}mV, Q_dc = {x0:.2f}mV",
            "\nAchieved LO supression:",
            f"{self.get_lo_leakage_rejection():.3f} dB",
        ]

        plt.text(
            0.05,
            0.75,
            "\n".join(content),
            horizontalalignment="left",
            verticalalignment="top",
            bbox=dict(facecolor="none", edgecolor="black", boxstyle="round,pad=1"),
        )
        plt.box(False)
        plt.xticks([])
        plt.yticks([])
        plt.suptitle(f"LO auto calibration @ {self.lo_frequency / 1e9:.3f}GHz")
        plt.tight_layout()

        return fig

    def show_image_rejection_calibration_result(
        self,
    ) -> matplotlib.figure.Figure:
        """
        Plots the results of the IF image calibration process.

        This method generates a series of subplots that visualize the coarse and fine scans of the IF image calibration
        and returns the achieved image suppression in dB when comparing the power without calibrating the mixers vs
        after calibration. The powers shown correspond to the power seen by the OPX+ inputs when the signal is routed
        internally through the Octave.

        Note: The power at the Octave RF output may vary up to 10dBm.

        Subplots:
        - Top Left: Summary of the current result and achieved image suppression.
        - Top Right: Coarse scan of gain vs phase with image power in dBm.
        - Bottom Left: Fit error in dBm.
        - Bottom Right: Fine scan of gain vs phase with image power in dBm.

        Returns:
            matplotlib.figure.Figure: The generated figure.
        """

        if_freq_data = self.if_data[self.if_frequency]

        fig = plt.figure(figsize=(11, 8.9))

        plt.subplot(222)
        r = if_freq_data.coarse

        zero_list = self._handle_zero_indices_and_masking(r.image)

        im = self.u.demod2volts(r.image, self.integration_length)
        im_dbm = self._convert_to_dbm(im)

        width = r.p_scan[0][1] - r.p_scan[0][0]
        height = r.g_scan[1][0] - r.g_scan[0][0]

        self._plot_scan(r.p_scan, r.g_scan, im_dbm, zero_list, width, height, "phase (rad)", "gain(%)")

        plt.plot(r.phase, r.gain, "yo", markersize=8)
        plt.plot(r.phase, r.gain, "ro", markersize=4)

        plt.text(
            r.phase,
            r.gain - 2 * height,
            f"phase={r.phase:.5f}\ngain={r.gain:.5f}",
            color="y",
            horizontalalignment="center",
            verticalalignment="top",
        )

        plt.text(
            np.min(r.p_scan) + 1.5 * width,
            np.max(r.g_scan - 1.5 * height),
            f"coarse scan\nLO = {self.lo_frequency / 1e9:.3f}GHz\nIF = {self.if_frequency / 1e6:.3f}MHz",
            color="k",
            bbox=dict(facecolor="w", alpha=0.8),
            verticalalignment="top",
        )

        r = if_freq_data.fine
        p_min = np.min(r.p_scan)
        p_max = np.max(r.p_scan)

        g_min = np.min(r.g_scan)
        g_max = np.max(r.g_scan)

        plt.plot(
            [p_min, p_max, p_max, p_min, p_min],
            [g_min, g_min, g_max, g_max, g_min],
            "w-.",
            linewidth=0.5,
        )

        plt.subplot(224)

        plt.xlabel("phase (rad)")
        plt.ylabel("gain(%)")
        plt.axis("equal")

        zero_list = self._handle_zero_indices_and_masking(r.image)

        im = self.u.demod2volts(r.image, self.integration_length)
        im_dbm = self._convert_to_dbm(im)

        width = r.p_scan[0][1] - r.p_scan[0][0]
        height = r.g_scan[1][0] - r.g_scan[0][0]

        self._plot_scan(r.p_scan, r.g_scan, im_dbm, zero_list, width, height, "phase (rad)", "gain(%)")

        plt.contour(r.p_scan, r.g_scan, im_dbm, colors="k", alpha=0.5)
        plt.plot(r.phase, r.gain, "yo", markersize=8)
        plt.plot(r.phase, r.gain, "ro", markersize=4)

        plt.text(
            r.phase,
            r.gain - 2 * height,
            f"phase={r.phase:.5f}\ngain={r.gain:.5f}",
            color="y",
            horizontalalignment="center",
            verticalalignment="top",
        )

        plt.text(
            np.min(r.p_scan) + 1.5 * width,
            np.max(r.g_scan - 1.5 * height),
            f"fine scan\nLO = {self.lo_frequency / 1e9:.3f}GHz\nIF = {self.if_frequency / 1e6:.3f}MHz",
            color="k",
            bbox=dict(facecolor="w", alpha=0.8),
            verticalalignment="top",
        )

        plt.subplot(223)

        X, Y = r.p_scan, r.g_scan
        p = r.fit.pol

        image = self.u.demod2volts(
            p[0] + p[1] * X + p[2] * Y + p[3] * X**2 + p[4] * X * Y + p[5] * Y**2 - r.image, self.integration_length
        )
        image_dbm = self._convert_to_dbm(np.abs(image))

        self._plot_scan(r.p_scan, r.g_scan, image_dbm, zero_list, width, height, "phase (rad)", "gain")

        plt.plot(r.phase, r.gain, "yo", markersize=8)
        plt.plot(r.phase, r.gain, "ro", markersize=4)

        # plt.colorbar(label="Power [dBm]")

        plt.text(
            np.min(r.p_scan) + 1.5 * width,
            np.max(r.g_scan - 1.5 * height),
            f"fit error\nLO = {self.lo_frequency / 1e9:.3f}GHz\nIF = {self.if_frequency / 1e6:.3f}MHz",
            color="k",
            bbox=dict(facecolor="w", alpha=0.8),
            verticalalignment="top",
        )

        plt.subplot(221)

        plt.xlim(0, 1)
        plt.ylim(0, 1)

        content = [
            "Calibrated parameters:",
            f"      gain = {r.gain * 100:.02f}%, phase = {r.phase * 180.0 / np.pi:.2f}deg",
            "\nAchieved Image sideband supression:",
            f"{self.get_image_rejection():.3f} dB",
        ]

        plt.text(
            0.05,
            0.75,
            "\n".join(content),
            horizontalalignment="left",
            verticalalignment="top",
            bbox=dict(facecolor="none", edgecolor="black", boxstyle="round,pad=1"),
        )
        plt.box(False)
        plt.xticks([])
        plt.yticks([])
        plt.suptitle(
            f"IMAGE auto calibration: LO = {self.lo_frequency / 1e9:.3f}GHz, IF = {self.if_frequency / 1e6:.3f}MHz"
        )
        plt.tight_layout()

        return fig

    def get_image_rejection(
        self,
    ):
        """
        Calculate the Image sideband suppression achieved by the automatic calibration.

        Returns:
            float: The reduction of Image sideband power without vs after calibration in dB units.
        """

        if_freq_data = self.if_data[self.if_frequency]
        image_fine = if_freq_data.fine.image

        _ = self._handle_zero_indices_and_masking(image_fine)
        image = self.u.demod2volts(image_fine, self.integration_length)
        image_array_dbm = self._convert_to_dbm(image)
        min_image_dbm = np.nanmin(image_array_dbm)

        pol = if_freq_data.fine.fit.pol
        image_0 = self._paraboloid(0, 0, pol)
        image_0_volts = self.u.demod2volts(image_0, self.integration_length)
        image_0_dbm = self._convert_to_dbm(image_0_volts)
        return min_image_dbm - image_0_dbm

    def get_lo_leakage_rejection(
        self,
    ):
        """
        Calculate the LO leakage suppression achieved by the automatic calibration.

        Returns:
            float: The reduction of LO leakage power without vs after calibration in dB units.
        """

        i_coarse = self.lo_data.debug.coarse[0].i_scan
        q_coarse = self.lo_data.debug.coarse[0].q_scan

        lo_coarse = self.lo_data.debug.coarse[0].lo
        lo_fine = self.lo_data.debug.fine[0].lo

        _ = self._handle_zero_indices_and_masking(lo_fine)
        lo = self.u.demod2volts(lo_fine, self.integration_length)
        lo_array_dbm = self._convert_to_dbm(lo)
        min_lo_dbm = np.nanmin(lo_array_dbm)

        id_i = np.argwhere(i_coarse[:, 0] == 0.0)
        id_q = np.argwhere(q_coarse[0, :] == 0.0)
        lo = lo_coarse[id_i, id_q][0][0]
        lo_0 = self.u.demod2volts(lo, self.integration_length)
        lo_0_dbm = self._convert_to_dbm(lo_0)

        return min_lo_dbm - lo_0_dbm

    @staticmethod
    def _paraboloid(x, y, pol):
        """
        Calculate the value of a paraboloid function.

        Args:
            x (float): The x-coordinate.
            y (float): The y-coordinate.
            pol (np.ndarray): The polynomial coefficients.

        Returns:
            float: The value of the paraboloid at (x, y).
        """
        return pol[0] + pol[1] * x + pol[2] * y + pol[3] * x**2 + pol[4] * x * y + pol[5] * y**2

    @staticmethod
    def _get_integration_length():
        # Get the source code of the function
        source_code = inspect.getsource(_add_calibration_entries_to_config)
        # Find the line where integration_length is defined
        for line in source_code.split("\n"):
            if "integration_length" in line:
                # Extract the value of integration_length
                integration_length = int(line.split("=")[1].strip())
                return integration_length
        return None



"""bakery.py: Framework to generate arbitrary waveforms to be played into QUA program
Author: Arthur Strauss - Quantum Machines
Created: 23/02/2021
"""

from typing import List, Union, Tuple, Dict, Optional, Set
from warnings import warn

import numpy as np
from qm import qua
import copy
from scipy.interpolate import interp1d


def baking(
    config,
    padding_method="right",
    override=False,
    baking_index: int = None,
    sampling_rate: Union[int, float] = 1e9,
):
    """
    Opens a context manager to synthesize samples for arbitrary waveforms. The input config is updated, unless a
    baking_index is given, in which case the generated waveforms can be retrieved using the get_waveform_dict() method,
    in a format readily pluggable as an overrides argument

    :param config: config file
    :param padding_method: Method to pad 0s to format the waveform to match hardware constraint
        (>16 ns and multiple of 4), can be set to "right", "left", "symmetric_r", "symmetric_l" or "none". If set to
        "none", an error is raised if the baked waveform does not match hardware constraints
    :param override: Define if baked waveforms are overridable when using add_compiled feature (default set to False)
    :param baking_index: index of a reference baking object to impose length constraint on new baked waveform
        (default set to None).
        If the reference waveforms do have different lengths for different quantum elements, the longest waveform
        is taken as a constraint.
        Moreover, if index is provided, input config is not updated with the waveform to be generated
        (useful for matching lengths when using waveform overriding in add_compiled feature)
    :param sampling_rate: Choose the number of samples per second (by default 1 Gs/sec) in
            which you write the waveforms within the baking. If the sampling rate higher than the default value,
            cubic interpolation is done to provide a new waveform at the OPX sampling rate. If the sampling rate is
            lower, the sampling rate argument is added to the waveform in the config, causing the OPX to interpolate
            in real time.


    """
    return Baking(config, padding_method, override, baking_index, sampling_rate)


class Baking:
    def __init__(
        self,
        config,
        padding_method: str = "right",
        override: bool = False,
        baking_index: int = None,
        sampling_rate: Union[int, float] = 1e9,
    ):
        self._config = config
        if baking_index is not None:
            self.update_config = False
        else:
            self.update_config = True
        self._padding_method = padding_method
        self._local_config = copy.deepcopy(config)
        self.sampling_rate = int(sampling_rate)
        (
            self._samples_dict,
            self._qe_dict,
            self._digital_samples_dict,
        ) = self._init_dict()
        self._ctr = self._find_baking_index(baking_index)  # unique name counter
        self._qe_set = set()
        self.override = override
        if override and sampling_rate < 1e9:
            raise ValueError(
                "Waveform can not be simultaneously overridable and compressed with lower than 1e9" "sampling rate"
            )
        self.length_constraint = self._retrieve_constraint_length(baking_index)
        self.override_waveforms_dict = {"waveforms": {}}
        self._out = True

    def __enter__(self):
        (
            self._samples_dict,
            self._qe_dict,
            self._digital_samples_dict,
        ) = self._init_dict()
        self._out = False
        return self

    @property
    def elements(self) -> Set:
        """
        Return the set of quantum elements involved in the baking
        """
        return self._qe_set

    @property
    def operations(self):
        """
        Access operations defined by the baking environment
        """
        return BakingOperations(self)

    @property
    def config(self) -> Dict:
        return self._config

    def _find_baking_index(self, baking_index: int = None) -> int:
        if baking_index is None:
            max_index = [-1]
            for qe in self._config["elements"].keys():
                index = [-1]
                for op in self._config["elements"][qe]["operations"]:
                    if op.find("baked") != -1:
                        index.append(int(op.split("_")[-1]))
                max_index.append(max(index))
            return max(max_index) + 1
        else:
            return baking_index

    def _init_dict(self):
        sample_dict = {}
        qe_dict = {}
        digit_samples_dict = {}
        for qe in self._config["elements"].keys():
            qe_dict[qe] = {
                "time": 0,
                "phase": 0.0,
                "freq": 0,
                "time_track": 0,  # Value used for negative waits, to know where to add the samples (negative int)
            }
            if any([key in self._local_config["elements"][qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                sample_dict[qe] = {"I": [], "Q": []}

            elif "singleInput" in self._local_config["elements"][qe]:
                sample_dict[qe] = {"single": []}
            digit_samples_dict[qe] = []

        return sample_dict, qe_dict, digit_samples_dict

    def _update_config(self, qe, qe_samples) -> None:
        # Generates new Op, pulse, and waveform for each qe to be added in the original config file
        self._config["elements"][qe]["operations"][f"baked_Op_{self._ctr}"] = f"{qe}_baked_pulse_{self._ctr}"
        if "I" in qe_samples:
            self._config["pulses"][f"{qe}_baked_pulse_{self._ctr}"] = {
                "operation": "control",
                "length": len(qe_samples["I"]),
                "waveforms": {
                    "I": f"{qe}_baked_wf_I_{self._ctr}",
                    "Q": f"{qe}_baked_wf_Q_{self._ctr}",
                },
            }
            self._config["waveforms"][f"{qe}_baked_wf_I_{self._ctr}"] = {
                "type": "arbitrary",
                "samples": qe_samples["I"],
                "is_overridable": self.override,
            }
            self._config["waveforms"][f"{qe}_baked_wf_Q_{self._ctr}"] = {
                "type": "arbitrary",
                "samples": qe_samples["Q"],
                "is_overridable": self.override,
            }
            if self.sampling_rate < int(1e9):
                self._config["waveforms"][f"{qe}_baked_wf_I_{self._ctr}"]["sampling_rate"] = self.sampling_rate
                self._config["waveforms"][f"{qe}_baked_wf_Q_{self._ctr}"]["sampling_rate"] = self.sampling_rate
        elif "single" in qe_samples:
            self._config["pulses"][f"{qe}_baked_pulse_{self._ctr}"] = {
                "operation": "control",
                "length": len(qe_samples["single"]),
                "waveforms": {"single": f"{qe}_baked_wf_{self._ctr}"},
            }
            self._config["waveforms"][f"{qe}_baked_wf_{self._ctr}"] = {
                "type": "arbitrary",
                "samples": qe_samples["single"],
                "is_overridable": self.override,
            }
            if self.sampling_rate < int(1e9):
                self._config["waveforms"][f"{qe}_baked_wf_{self._ctr}"]["sampling_rate"] = self.sampling_rate

        if len(self._digital_samples_dict[qe]) != 0:
            self._config["pulses"][f"{qe}_baked_pulse_{self._ctr}"][
                "digital_marker"
            ] = f"{qe}_baked_digital_wf_{self._ctr}"
            if "digital_waveforms" in self._config:
                self._config["digital_waveforms"][f"{qe}_baked_digital_wf_{self._ctr}"] = {
                    "samples": self._digital_samples_dict[qe]
                }
            else:
                self._config["digital_waveforms"] = {}
                self._config["digital_waveforms"][f"{qe}_baked_digital_wf_{self._ctr}"] = {
                    "samples": self._digital_samples_dict[qe]
                }

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Updates the configuration dictionary upon exit
        """
        if exc_type:
            return
        self._out = True
        elements = self._local_config["elements"]
        for qe in elements:
            wait_duration = 0  # Stores the duration that has to be padded with 0s to make a valid sample for QUA
            # in original config file

            if self._qe_dict[qe]["time"] > 0:  # Check if sample was added to the quantum element
                # otherwise we do not add any Op
                self._qe_set.add(qe)
                qe_samples = self._samples_dict[qe]
                if self.sampling_rate > int(1e9):
                    if any([key in elements[qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                        y1 = qe_samples["I"]
                        y2 = qe_samples["Q"]
                        dt = 1e9 / self.sampling_rate
                        max_t = len(y1) * dt
                        x = np.arange(0, max_t - dt / 2, dt)
                        f1 = interp1d(x, y1, "cubic")
                        f2 = interp1d(x, y2, "cubic")
                        x_new = np.arange(0, int(max_t))
                        self._qe_dict[qe]["time"] = int(max_t)
                        qe_samples["I"] = list(f1(x_new))
                        qe_samples["Q"] = list(f2(x_new))
                    elif "singleInput" in elements[qe]:
                        y = qe_samples["single"]
                        dt = 1e9 / self.sampling_rate
                        max_t = len(y) * dt
                        x = np.arange(0, max_t - dt / 2, dt)
                        f = interp1d(x, y, "cubic")
                        x_new = np.arange(0, int(max_t))
                        self._qe_dict[qe]["time"] = int(max_t)
                        qe_samples["single"] = list(f(x_new))

                if self.length_constraint is not None:
                    assert self._qe_dict[qe]["time"] <= self.length_constraint, (
                        f"Provided length constraint (={self.length_constraint}) "
                        f"smaller than actual baked samples length ({self._qe_dict[qe]['time']})"
                    )
                    wait_duration += self.length_constraint - self._qe_dict[qe]["time"]
                    self.wait(self.length_constraint - self._qe_dict[qe]["time"], qe)
                if self._qe_dict[qe]["time"] < 16:  # Sample length must be at least 16 ns long
                    wait_duration += 16 - self._qe_dict[qe]["time"]
                    self.wait(16 - self._qe_dict[qe]["time"], qe)
                if not self._qe_dict[qe]["time"] % 4 == 0:  # Sample length must be a multiple of 4
                    wait_duration += 4 - self._qe_dict[qe]["time"] % 4
                    self.wait(4 - self._qe_dict[qe]["time"] % 4, qe)

                end_samples = 0
                if any([key in elements[qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                    end_samples = len(qe_samples["I"]) - wait_duration
                elif "singleInput" in elements[qe]:
                    end_samples = len(qe_samples["single"]) - wait_duration

                # Padding done according to desired method, can be either right, left, symmetric left or symmetric right
                if self._padding_method == "right":
                    pass
                elif self._padding_method == "none":
                    wait_duration2 = wait_duration
                    if self.length_constraint is not None:
                        wait_duration2 -= self.length_constraint
                    if wait_duration2 != 0:
                        raise ValueError(
                            "Baked waveform requires padding to match hardware constraint"
                            " whereas no padding is desired"
                        )

                elif self._padding_method == "left":
                    if any([key in elements[qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                        qe_samples["I"] = qe_samples["I"][end_samples:] + qe_samples["I"][0:end_samples]
                        qe_samples["Q"] = qe_samples["Q"][end_samples:] + qe_samples["Q"][0:end_samples]
                    elif "singleInput" in elements[qe]:
                        qe_samples["single"] = qe_samples["single"][end_samples:] + qe_samples["single"][0:end_samples]

                elif self._padding_method == "symmetric_l" or (
                    self._padding_method == "symmetric_r" and wait_duration % 2 == 0
                ):
                    if any([key in elements[qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                        qe_samples["I"] = (
                            qe_samples["I"][end_samples + wait_duration // 2 :]
                            + qe_samples["I"][0 : end_samples + wait_duration // 2]
                        )

                        qe_samples["Q"] = (
                            qe_samples["Q"][end_samples + wait_duration // 2 :]
                            + qe_samples["Q"][0 : end_samples + wait_duration // 2]
                        )

                    elif "singleInput" in elements[qe]:
                        qe_samples["single"] = (
                            qe_samples["single"][end_samples + wait_duration // 2 :]
                            + qe_samples["single"][0 : end_samples + wait_duration // 2]
                        )

                elif self._padding_method == "symmetric_r" and wait_duration % 2 != 0:
                    if any([key in elements[qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                        qe_samples["I"] = (
                            qe_samples["I"][end_samples + wait_duration // 2 + 1 :]
                            + qe_samples["I"][0 : end_samples + wait_duration // 2 + 1]
                        )
                        qe_samples["Q"] = (
                            qe_samples["Q"][end_samples + wait_duration // 2 + 1 :]
                            + qe_samples["Q"][0 : end_samples + wait_duration // 2 + 1]
                        )
                    elif "singleInput" in elements[qe]:
                        qe_samples["single"] = (
                            qe_samples["single"][end_samples + wait_duration // 2 + 1 :]
                            + qe_samples["single"][0 : end_samples + wait_duration // 2 + 1]
                        )

                if self.update_config:
                    self._update_config(qe, qe_samples)

                if any([key in elements[qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                    self.override_waveforms_dict["waveforms"][f"{qe}_baked_wf_I_{self._ctr}"] = qe_samples["I"]
                    self.override_waveforms_dict["waveforms"][f"{qe}_baked_wf_Q_{self._ctr}"] = qe_samples["Q"]

                elif "singleInput" in elements[qe]:
                    self.override_waveforms_dict["waveforms"][f"{qe}_baked_wf_{self._ctr}"] = qe_samples["single"]

    def is_out(self):
        return self._out

    def _get_samples(self, pulse: str) -> Union[List[float], List[List[float]]]:
        """
        Returns samples associated with a pulse
        :param pulse:
        :returns: Python list containing samples, [samples_I, samples_Q] in case of mixInputs
        """

        try:
            waveforms = self._local_config["waveforms"]
            pulses = self._local_config["pulses"]

            if self.sampling_rate > int(1e9) and pulse in self._config["pulses"]:
                dt = 1e9 / self.sampling_rate
                orig_max_t = pulses[pulse]["length"]
                len_t = np.ceil(orig_max_t / dt).astype(int)
                if "single" in self._local_config["pulses"][pulse]["waveforms"]:
                    wf = pulses[pulse]["waveforms"]["single"]
                    if waveforms[wf]["type"] == "constant":
                        return [waveforms[wf]["sample"]] * len_t
                    else:
                        y = waveforms[wf]["samples"]
                        x = np.arange(0, len(y))
                        f = interp1d(x, y, "cubic", fill_value="extrapolate")
                        x_new = np.arange(0, orig_max_t, dt)
                        return list(f(x_new))

                elif "I" in pulses[pulse]["waveforms"]:
                    wf_I = pulses[pulse]["waveforms"]["I"]
                    wf_Q = pulses[pulse]["waveforms"]["Q"]
                    if waveforms[wf_I]["type"] == "constant":
                        samples_I = [waveforms[wf_I]["sample"]] * len_t
                    else:
                        y1 = waveforms[wf_I]["samples"]
                        x1 = np.arange(0, len(y1))
                        f1 = interp1d(x1, y1, "cubic", fill_value="extrapolate")
                        x_new1 = np.arange(0, orig_max_t, dt)
                        samples_I = list(f1(x_new1))

                    if waveforms[wf_Q]["type"] == "constant":
                        samples_Q = [waveforms[wf_Q]["sample"]] * len_t
                    else:
                        y2 = waveforms[wf_Q]["samples"]
                        x2 = np.arange(0, len(y2))
                        f2 = interp1d(x2, y2, "cubic", fill_value="extrapolate")
                        x_new2 = np.arange(0, orig_max_t, dt)
                        samples_Q = list(f2(x_new2))
                    return [samples_I, samples_Q]

            else:
                if "single" in pulses[pulse]["waveforms"]:
                    wf = pulses[pulse]["waveforms"]["single"]
                    if waveforms[wf]["type"] == "constant":
                        return [waveforms[wf]["sample"]] * pulses[pulse]["length"]
                    else:
                        return list(waveforms[wf]["samples"])
                elif "I" in pulses[pulse]["waveforms"]:
                    wf_I = pulses[pulse]["waveforms"]["I"]
                    wf_Q = pulses[pulse]["waveforms"]["Q"]
                    if waveforms[wf_I]["type"] == "constant":
                        samples_I = [waveforms[wf_I]["sample"]] * pulses[pulse]["length"]
                    else:
                        samples_I = list(waveforms[wf_I]["samples"])
                    if waveforms[wf_Q]["type"] == "constant":
                        samples_Q = [waveforms[wf_Q]["sample"]] * pulses[pulse]["length"]
                    else:
                        samples_Q = list(waveforms[wf_Q]["samples"])
                    return [samples_I, samples_Q]

        except KeyError:
            raise KeyError(f"No waveforms found for pulse {pulse}")

    def get_baking_index(self) -> int:
        """
        :return: Index of the baking object (based on its order of creation)
        """
        return self._ctr

    def get_current_length(self, qe: Optional[str] = None) -> int:
        """
        Retrieve within the baking the current length of the waveform being created (within the baking)

        :param qe: quantum element
        """
        if qe is None:
            max_length = 0
            for qe in self._qe_dict:
                if self._qe_dict[qe]["time"] > 0:
                    length = self.get_op_length(qe)
                    if length > max_length:
                        max_length = length
            return max_length
        else:
            if any([key in self._local_config["elements"][qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                return len(self._samples_dict[qe]["I"])
            elif "singleInput" in self._local_config["elements"][qe]:
                return len(self._samples_dict[qe]["single"])
            else:
                raise KeyError("Element not in the config or does not have any analog input.")

    def _get_pulse_index(self, qe) -> int:
        index = 0
        for pulse in self._local_config["pulses"]:
            if pulse.find(f"{qe}_baked_pulse_b{self._ctr}") != -1:
                index += 1
        return index

    def get_qe_set(self):
        if self._out:
            return self._qe_set
        else:
            qe_set = set()
            for qe in self._qe_dict:
                if self._qe_dict[qe]["time"] > 0:
                    qe_set.add(qe)
            return qe_set

    def get_waveforms_dict(self) -> Dict:
        """

        :return: Dictionary of baked waveforms for all quantum elements, in a format compatible
            with overrides argument of the add_compiled method of QM API
        """
        return self.override_waveforms_dict

    def delete_samples(
        self,
        t_start: int,
        t_stop: Optional[int] = None,
        qe: Optional[Union[List[str], str]] = None,
    ):
        """
        Delete samples from the current list of samples composing the baked waveform

        :param t_start: time index from where to start the deletion. If t_start < 0, samples are deleted from the end of
            the current waveform
        :param t_stop: time index to where to end the deletion. t_stop must be higher than t_start (t_stop > t_start).
            If t_start is negative, t_stop is not taken into account
        :param qe: Element or list of quantum elements for which to delete the samples. If None is provided, baked
        samples are deleted for all quantum elements involved in the baking object
        """

        def delete_for_el(qe_internal, t_start_internal, t_stop_internal):
            input_type = []

            if self._qe_dict[qe_internal]["time"] > 0:
                if any(
                    [
                        key in self._local_config["elements"][qe_internal]
                        for key in ["mixInputs", "RF_inputs", "MWInput"]
                    ]
                ):
                    input_type.append("I")
                    input_type.append("Q")
                    ref_length = len(self._samples_dict[qe_internal]["I"])
                elif "singleInput" in self._local_config["elements"][qe_internal]:
                    input_type.append("single")
                    ref_length = len(self._samples_dict[qe_internal]["single"])
                else:
                    raise ValueError("Element provided does not have any analog input")
                if t_start_internal < 0:
                    if t_start_internal < ref_length:
                        self._update_qe_time(qe_internal, t_start_internal)
                        for i in input_type:
                            del self._samples_dict[qe_internal][i][t_start_internal:]
                    else:
                        raise ValueError("Desired deletion exceeds current waveform length")
                else:
                    if t_stop_internal is None:
                        t_stop_internal = ref_length
                    if t_stop_internal > ref_length:
                        raise ValueError("Desired deletion exceeds current waveform length")
                    if t_start_internal > ref_length:
                        raise ValueError("t_start is higher than current waveform length")
                    if t_start_internal > t_stop_internal:
                        raise ValueError("t_stop is smaller than t_start")

                    self._update_qe_time(qe_internal, t_start_internal - t_stop_internal)
                    for i in input_type:
                        del self._samples_dict[qe_internal][i][t_start_internal:t_stop_internal]

        if not self._out:
            if qe is not None:
                if isinstance(qe, List):
                    for q in qe:
                        delete_for_el(q, t_start, t_stop)
                elif isinstance(qe, str):
                    delete_for_el(qe, t_start, t_stop)
            else:
                for q in self._qe_dict:
                    delete_for_el(q, t_start, t_stop)
        else:
            raise Warning(
                "Cannot delete samples whe outside of the baking context manager, use delete_baked_Op " "instead"
            )

    def delete_baked_op(self, *qe_set: str) -> None:
        """
        Delete from the config the baked operation and its associated pulse and waveform(s) for the
        specified quantum_elements

        :param qe_set:
            tuple of quantum elements, if no element is provided,
            all the baked operations associated to this baking object will be deleted
        """

        def remove_op(q):
            if self.length_constraint is None:
                if self._out:
                    if f"baked_Op_{self._ctr}" in self.config["elements"][q]["operations"]:
                        del self.config["elements"][q]["operations"][f"baked_Op_{self._ctr}"]
                        del self.config["pulses"][f"{q}_baked_pulse_{self._ctr}"]
                        if any([key in self.config["elements"][q] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                            del self.config["waveforms"][f"{q}_baked_wf_I_{self._ctr}"]
                            del self.config["waveforms"][f"{q}_baked_wf_Q_{self._ctr}"]
                        elif "singleInput" in self.config["elements"][q]:
                            del self.config["waveforms"][f"{q}_baked_wf_{self._ctr}"]
                        if "digital_waveforms" in self._config:
                            if f"{q}_baked_digital_wf_{self._ctr}" in self._config["digital_waveforms"]:
                                del self.config["digital_waveforms"][f"{q}_baked_digital_wf_{self._ctr}"]

                else:
                    raise KeyError(
                        "delete_baked_Op only available outside of the context manager "
                        "(config is updated at the exit)"
                    )
            else:
                raise Warning("Operation could not be deleted because baking object does not update the config")

        if len(qe_set) != 0:
            for qe in qe_set:
                remove_op(qe)
        else:
            for qe in self._qe_dict.keys():
                remove_op(qe)

    def get_op_name(self, qe: str) -> str:
        """
        Get the baked operation issued from the baking object for quantum element qe

        :param qe: quantum element carrying the baked operation
        :returns: Name of baked operation associated to element qe
        """
        if not (qe in self._qe_set):
            raise KeyError(f"{qe} is not in the set of quantum elements of the baking object ")
        else:
            return f"baked_Op_{self._ctr}"

    def get_op_length(self, qe: Optional[str] = None) -> int:
        """
        Retrieve the length of the finalized baked waveform associated to quantum element qe (outside the baking)

        :param qe: target quantum element, if None is provided, then length of the longest baked waveform
            associated to this baking object is returned
        :returns: Length of baked operation associated to element qe (or maximum length if None is provided)

        """
        if qe is not None:
            if self.update_config and self._out:
                if not (qe in self._qe_set):
                    raise KeyError(f"{qe} is not in the set of quantum elements of the baking object ")
                else:
                    if any([key in self._config["elements"][qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                        return len(self._config["waveforms"][f"{qe}_baked_wf_I_{self._ctr}"]["samples"])
                    else:
                        return len(self._config["waveforms"][f"{qe}_baked_wf_{self._ctr}"]["samples"])
            else:
                return self.get_current_length(qe)

        else:
            max_length = 0
            for qe in self._qe_dict:
                if self._qe_dict[qe]["time"] > 0:
                    length = self.get_op_length(qe)
                    if length > max_length:
                        max_length = length
            return max_length

    def add_digital_waveform(self, name: str, digital_samples: List[Tuple]) -> None:
        """
        Adds a digital waveform to be attached to a baked operation created using the add_op method

        :param name: name of the digital waveform
        :param digital_samples: samples used to generate digital_waveform
        """
        if "digital_waveforms" in self._local_config:
            self._local_config["digital_waveforms"][name] = {"samples": digital_samples}
        else:
            self._local_config["digital_waveforms"] = {}
            self._local_config["digital_waveforms"][name] = {"samples": digital_samples}

    def add_Op(
        self,
        name: str,
        qe: str,
        samples: Union[List[float], List[List[float]]],
        digital_marker: str = None,
    ) -> None:
        warn(
            "This method is deprecated, use add_op() instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.add_op(name, qe, samples, digital_marker)

    def add_op(
        self,
        name: str,
        qe: str,
        samples: Union[List[float], List[List[float]]],
        digital_marker: str = None,
    ) -> None:
        """
        Adds an operation playable within the baking context manager.

        :param name: name of the Operation to be added for the quantum element (to be used only within the context manager)
        :param qe: targeted quantum element
        :param  samples: arbitrary waveforms to be inserted into pulse definition. Should be in the same units as the
        sampling rate used by the baking context manager (default 1 Gs/sec)
        :param digital_marker: name of the digital marker sample associated to the generated pulse (assumed to be in the original config)
        """

        index = self._get_pulse_index(qe)
        Op = {name: f"{qe}_baked_pulse_b{self._ctr}_{index}"}
        if any([key in self._local_config["elements"][qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
            assert len(samples) == 2, f"{qe} is a mixInputs/RF_inputs element, two lists should be provided"
            assert len(samples[0]) == len(
                samples[1]
            ), "Error : samples provided for I and Q do not have the same length"

            pulse = {
                f"{qe}_baked_pulse_b{self._ctr}_{index}": {
                    "operation": "control",
                    "length": len(samples[0]),
                    "waveforms": {
                        "I": f"{qe}_baked_b{self._ctr}_{index}_wf_I",
                        "Q": f"{qe}_baked_b{self._ctr}_{index}_wf_Q",
                    },
                }
            }
            if digital_marker is not None:
                pulse[f"{qe}_baked_pulse_b{self._ctr}_{index}"]["digital_marker"] = digital_marker

            waveform = {
                f"{qe}_baked_b{self._ctr}_{index}_wf_I": {
                    "type": "arbitrary",
                    "samples": samples[0],
                },
                f"{qe}_baked_b{self._ctr}_{index}_wf_Q": {
                    "type": "arbitrary",
                    "samples": samples[1],
                },
            }

        elif "singleInput" in self._local_config["elements"][qe]:
            for i in range(len(samples)):
                assert (
                    type(samples[i]) is float or type(samples[i]) is int
                ), f"{qe} is a singleInput element, list of numbers (int or float) should be provided "

            pulse = {
                f"{qe}_baked_pulse_b{self._ctr}_{index}": {
                    "operation": "control",
                    "length": len(samples),
                    "waveforms": {"single": f"{qe}_baked_b{self._ctr}_{index}_wf"},
                }
            }

            if digital_marker is not None:
                pulse[f"{qe}_baked_pulse_b{self._ctr}_{index}"]["digital_marker"] = digital_marker
            waveform = {
                f"{qe}_baked_b{self._ctr}_{index}_wf": {
                    "type": "arbitrary",
                    "samples": samples,
                }
            }
        else:
            raise ValueError("Element provided does not have any analog input")
        self._local_config["pulses"].update(pulse)
        self._local_config["waveforms"].update(waveform)
        self._local_config["elements"][qe]["operations"].update(Op)

    def play(self, Op: str, qe: str, amp: Union[float, Tuple[float]] = 1.0) -> None:
        """
        Add a pulse to the baked sequence

        :param Op: operation to play to quantum element
        :param qe: targeted quantum element
        :param amp:
            amplitude of the pulse, can be either a float or a tuple of 4 variables
            (similar to amp(a) or amp(v00, v01, v10, v11) in QUA)
        """
        try:
            pulse = self._local_config["elements"][qe]["operations"][Op]
            samples = self._get_samples(pulse)
            freq = self._qe_dict[qe]["freq"]
            phi = self._qe_dict[qe]["phase"]

            if self._qe_dict[qe]["time_track"] == 0:
                if any([key in self._local_config["elements"][qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                    assert isinstance(
                        samples, list
                    ), f"{qe} is a mixInputs/RF_inputs element, two lists should be provided"
                    assert len(samples) == 2, f"{qe} is a mixInputs/RF_inputs element, two lists should be provided"
                    assert type(samples[0] == list) and type(
                        samples[1] == list
                    ), f"{qe} is a mixInputs/RF_inputs element, two lists should be provided"

                    assert len(samples[0]) == len(samples[1]), (
                        f"Error : samples provided for I and Q do not have the same length. length I: {len(samples[0])}"
                        f", length Q: {len(samples[1])}"
                    )

                    I = samples[0]
                    Q = samples[1]
                    I2 = [None] * len(I)
                    Q2 = [None] * len(Q)
                    I3 = [None] * len(I)
                    Q3 = [None] * len(Q)

                    for i in range(len(I)):
                        if type(amp) is float or type(amp) is int:
                            I2[i] = amp * I[i]
                            Q2[i] = amp * Q[i]
                        elif len(amp) != 4 or type(amp) is not tuple:
                            raise IndexError("Amplitudes provided must be stored in a tuple (v00, v01, v10, v11)")
                        else:
                            I2[i] = amp[0] * I[i] + amp[1] * Q[i]
                            Q2[i] = amp[2] * I[i] + amp[3] * Q[i]
                        I3[i] = np.cos(freq * i * 1e-9 + phi) * I2[i] - np.sin(freq * i * 1e-9 + phi) * Q2[i]
                        Q3[i] = np.sin(freq * i * 1e-9 + phi) * I2[i] + np.cos(freq * i * 1e-9 + phi) * Q2[i]
                        self._samples_dict[qe]["I"].append(I3[i])
                        self._samples_dict[qe]["Q"].append(Q3[i])
                    self._update_qe_time(qe, len(I))

                elif "singleInput" in self._local_config["elements"][qe]:
                    for i in range(len(samples)):
                        assert isinstance(
                            samples[i], (float, int)
                        ), f"{qe} is a singleInput element, list of numbers (int or float) should be provided "
                        self._samples_dict[qe]["single"].append(amp * np.cos(freq * i * 1e-9 + phi) * samples[i])
                    self._update_qe_time(qe, len(samples))

                else:
                    raise ValueError("Element provided does not have any analog input")
                # Update of digital waveform
                if "digital_marker" in self._local_config["pulses"][pulse]:
                    digital_marker = self._local_config["pulses"][pulse]["digital_marker"]
                    digital_waveform = self._local_config["digital_waveforms"][digital_marker]["samples"]
                    self._digital_samples_dict[qe] += digital_waveform

            else:
                self.play_at(Op, qe, self._qe_dict[qe]["time_track"], amp)
                self._qe_dict[qe]["time_track"] = 0

        except KeyError:
            raise KeyError(f'Op:"{Op}" does not exist in configuration and not manually added (use add_pulse)')

    def play_at(self, Op: str, qe: str, t: int, amp: Union[float, Tuple[float]] = 1.0) -> None:
        """
        Add a waveform to the sequence at the specified time index.
        If t is higher than pulse duration for the specified quantum element,
        a wait command is inserted until the indicated time (in ns).
        Otherwise, waveform is added (addition of samples) to the pre-existing sequence.
        Finally, providing a negative index starts adding the sample with a prior negative wait of t
        Note that the phase played for the newly formed sample is the one that was set before adding the new waveform

        :param Op: operation to play to quantum element
        :param qe: targeted quantum element
        :param t: Time tag in ns where the pulse should be added
        :param amp: amplitude of the pulse, can be either a float or a tuple of 4 variables (similar to amp(a) or amp(v00, v01, v10, v11) in QUA)
        """
        freq = self._qe_dict[qe]["freq"]
        phi = self._qe_dict[qe]["phase"]
        if type(t) is not int:
            if type(t) is float:
                t = int(t)
            else:
                raise TypeError("Provided time is not an integer")
        elif t < 0:
            self.wait(t, qe)  # Negative wait
            self.play(Op, qe, amp)
        elif t > self._qe_dict[qe]["time"]:
            self.wait(t - self._qe_dict[qe]["time"], qe)
            self.play(Op, qe, amp)
        else:
            try:
                pulse = self._local_config["elements"][qe]["operations"][Op]
                samples = self._get_samples(pulse)
                new_samples = 0
                if any([key in self._local_config["elements"][qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
                    assert isinstance(
                        samples, list
                    ), f"{qe} is a mixInputs/RF_inputs element, two lists should be provided"
                    assert len(samples) == 2, f"{qe} is a mixInputs/RF_inputs element, two lists should be provided"
                    assert type(samples[0] == list) and type(samples[1] == list), (
                        f"{qe} is a mixInputs/RF_inputs element, " f"two lists should be provided"
                    )

                    assert len(samples[0]) == len(
                        samples[1]
                    ), "Error : samples provided for I and Q do not have the same length"

                    I, Q = samples[0], samples[1]
                    I2, Q2, I3, Q3 = (
                        [None] * len(I),
                        [None] * len(I),
                        [None] * len(I),
                        [None] * len(I),
                    )
                    for i in range(len(I)):
                        if type(amp) is float or type(amp) is int:
                            I2[i] = amp * I[i]
                            Q2[i] = amp * Q[i]
                        else:
                            if len(amp) != 4 or type(amp) is not tuple:
                                raise IndexError("Amplitudes provided must be stored in a tuple (v00, v01, v10, v11)")
                            else:
                                I2[i] = amp[0] * I[i] + amp[1] * Q[i]
                                Q2[i] = amp[2] * I[i] + amp[3] * Q[i]
                        if t + i < len(self._samples_dict[qe]["I"]):
                            I3[i] = (
                                np.cos(freq * (t + i) * 1e-9 + phi) * I2[i]
                                - np.sin(freq * (t + i) * 1e-9 + phi) * Q2[i]
                            )
                            Q3[i] = (
                                np.sin(freq * (t + i) * 1e-9 + phi) * I2[i]
                                + np.cos(freq * (t + i) * 1e-9 + phi) * Q2[i]
                            )

                            self._samples_dict[qe]["I"][t + i] += I3[i]
                            self._samples_dict[qe]["Q"][t + i] += Q3[i]
                        else:
                            I3[i] = np.cos(freq * i * 1e-9 + phi) * I2[i] - np.sin(freq * i * 1e-9 + phi) * Q2[i]
                            Q3[i] = np.sin(freq * i * 1e-9 + phi) * I2[i] + np.cos(freq * i * 1e-9 + phi) * Q2[i]

                            self._samples_dict[qe]["I"].append(I3[i])
                            self._samples_dict[qe]["Q"].append(Q3[i])
                            new_samples += 1

                elif "singleInput" in self._local_config["elements"][qe]:
                    if type(amp) is not float and type(amp) is not int:
                        raise IndexError("Amplitude must be a number")

                    for i in range(len(samples)):
                        assert (
                            type(samples[i]) is float or type(samples[i]) is int
                        ), f"{qe} is a singleInput element, list of numbers (int or float) should be provided "
                        if t + i < len(self._samples_dict[qe]["single"]):
                            self._samples_dict[qe]["single"][t + i] += (
                                amp * np.cos(freq * (t + i) * 1e-9 + phi) * samples[i]
                            )
                        else:
                            self._samples_dict[qe]["single"].append(amp * samples[i])
                            new_samples += 1
                else:
                    raise ValueError("Element provided does not have any analog input")

                self._update_qe_time(qe, new_samples)

            except KeyError:
                raise KeyError(f'Op:"{Op}" does not exist in configuration and not manually added (use add_pulse)')

    def frame_rotation(self, angle: float, qe: str) -> None:
        """
        Shift the phase of the oscillator associated with a quantum element by the given angle.
        This is typically used for virtual z-rotations.
        Frame rotation done within the baking sticks to the rest of the
        QUA program after its execution.

        :param angle: phase parameter
        :param qe: quantum element
        """

        self._update_qe_phase(qe, angle)

    def frame_rotation_2pi(self, angle: float, qe: str) -> None:
        """
        Shift the phase of the oscillator associated with a quantum element by the given angle.
        This is typically used for virtual z-rotations.
        Frame rotation done within the baking sticks to the rest of the
        QUA program after its execution. This performs a frame rotation of 2*π*angle

        :param angle: phase parameter
        :param qe: quantum element
        """

        self._update_qe_phase(qe, 2 * np.pi * angle)

    def set_detuning(self, qe: str, freq: int) -> None:
        """Update frequency by adding detuning to original IF set in the config.
        Unlike frame rotation, the detuning will only affect the baked operation and will not stick to the element

        :param qe: quantum element
        :param freq: frequency of the detuning (in Hz)
        """
        self._qe_dict[qe]["freq"] = freq

    def reset_frame(self, *qe_set: str) -> None:
        """
        Used to reset all the frame updated made up to this statement.

        .. note::
            This is not equivalent to the QUA reset_frame() command and will only remove the phase updated within the
            baking

        :param qe_set: Set[str] of quantum elements
        """
        for qe in qe_set:
            self._update_qe_phase(qe, -self._qe_dict[qe]["phase"])

    def ramp(self, amp: float, duration: int, qe: str) -> None:
        """
        Analog of ramp function in QUA

        :param amp: slope
        :param duration: duration of ramping
        :param qe: quantum element
        """
        ramp_sample = [amp * t for t in range(duration)]
        if "singleInput" in self._local_config["elements"][qe]:
            self._samples_dict[qe]["single"] += ramp_sample
        elif any([key in self._local_config["elements"][qe] for key in ["mixInputs", "RF_inputs", "MWInput"]]):
            self._samples_dict[qe]["Q"] += ramp_sample
            self._samples_dict[qe]["I"] += [0] * duration
        self._update_qe_time(qe, duration)

    def _update_qe_time(self, qe: str, dt: int) -> None:
        self._qe_dict[qe]["time"] += dt

    def _update_qe_phase(self, qe: str, phi: float) -> None:
        self._qe_dict[qe]["phase"] += phi

    def wait(self, duration: int, *qe_set: str) -> None:
        """
        Wait for the given duration on all provided elements.
        Here, the wait is simply adding 0 to the existing sample for a given duration.

        :param duration: waiting duration
        :param qe_set: set of quantum elements

        """
        if duration >= 0:
            for qe in qe_set:
                if qe in self._samples_dict.keys():
                    if any(
                        [
                            key in self._local_config["elements"][qe].keys()
                            for key in ["mixInputs", "RF_inputs", "MWInput"]
                        ]
                    ):
                        self._samples_dict[qe]["I"] += [0] * duration
                        self._samples_dict[qe]["Q"] += [0] * duration

                    elif "singleInput" in self._local_config["elements"][qe].keys():
                        self._samples_dict[qe]["single"] += [0] * duration

                self._update_qe_time(qe, duration)

        else:
            for qe in qe_set:
                # Duration is negative so just add for subtraction
                self._qe_dict[qe]["time_track"] = self._qe_dict[qe]["time"] + duration
                if self._qe_dict[qe]["time_track"] < 0:
                    raise ValueError(
                        f"Negative wait chosen (= {duration}) too large for current baked samples length ("
                        f"= {self.get_current_length(qe)})"
                    )

    def align(self, *qe_set: str) -> None:
        """
        Align several quantum elements together.
        All the quantum elements referenced in *elements will wait for all
        the others to finish their currently running statement.

        :param qe_set: set of quantum elements to be aligned altogether, if no element is passed, alignment is done
            for all elements within the baking
        """

        def alignment(qe_set2):
            last_qe = ""
            last_t = 0
            for qe in qe_set2:
                qe_t = self._qe_dict[qe]["time"]
                if qe_t > last_t:
                    last_qe = qe
                    last_t = qe_t

            for qe in qe_set2:
                qe_t = self._qe_dict[qe]["time"]
                if qe != last_qe:
                    self.wait(last_t - qe_t, qe)

        if len(qe_set) == 0:
            alignment(self.get_qe_set())
        else:
            alignment(qe_set)

    def run(
        self,
        amp_array: List[Tuple] = None,
        trunc_array: List[Tuple] = None,
        align_elements: bool = True,
    ) -> None:
        """
        Plays the baked waveform
        This method must be used within a QUA program

        :param amp_array: list of tuples for amplitudes (e.g [(qe1, amp1), (qe2, amp2)] ), each amplitude must be a scalar
        :param trunc_array: list of tuples for truncations (e.g [(qe1, amp1), (qe2, amp2)] ), each truncation must be an
            int or QUA int
        :param align_elements: If true (default) and there are more than one element in the baking, adds an align
            command between the elements.
        """

        qe_set = self.get_qe_set()

        if align_elements:
            if len(qe_set) > 1:
                qua.align(*qe_set)

        if trunc_array is None:
            if amp_array is None:
                for qe in qe_set:
                    qua.play(f"baked_Op_{self._ctr}", qe)

            else:
                for qe in qe_set:
                    if not (qe in list(zip(*amp_array))[0]):
                        qua.play(f"baked_Op_{self._ctr}", qe)

                    else:
                        index2 = list(zip(*amp_array))[0].index(qe)
                        amp = list(zip(*amp_array))[1][index2]
                        if type(amp) is list:
                            raise TypeError("Amplitude can only be a number (either Python or QUA variable)")
                        qua.play(f"baked_Op_{self._ctr}" * qua.amp(amp), qe)

        else:
            for qe in qe_set:
                if qe not in list(zip(*trunc_array))[0]:
                    trunc_array.append((qe, None))
                index = list(zip(*trunc_array))[0].index(qe)
                trunc = list(zip(*trunc_array))[1][index]
                if amp_array is None:
                    qua.play(f"baked_Op_{self._ctr}", qe, truncate=trunc)

                else:
                    if not (qe in list(zip(*amp_array))[0]):
                        qua.play(f"baked_Op_{self._ctr}", qe, truncate=trunc)

                    else:
                        index2 = list(zip(*amp_array))[0].index(qe)
                        amp = list(zip(*amp_array))[1][index2]
                        if type(amp) is list:
                            raise TypeError("Amplitude can only be a number (either Python or QUA variable)")
                        qua.play(f"baked_Op_{self._ctr}" * qua.amp(amp), qe, truncate=trunc)

        for qe in qe_set:
            if self._qe_dict[qe]["phase"] != 0:
                qua.frame_rotation_2pi(self._qe_dict[qe]["phase"] / (2 * np.pi), qe)

    def _retrieve_constraint_length(self, baking_index: int = None) -> Optional[int]:
        if baking_index is not None:
            max_length = 0
            for pulse in self._local_config["pulses"]:
                if (
                    pulse.find(f"baked_pulse_{baking_index}") != -1
                    and self._local_config["pulses"][pulse]["length"] > max_length
                ):
                    max_length = self._local_config["pulses"][pulse]["length"]

            return max_length
        else:
            return None


def deterministic_run(baking_list, j, unsafe=False) -> None:
    """
    Generates a QUA macro for a binary tree ensuring a synchronized play of operations
    listed in the various baking objects

    :param baking_list: Python list of Baking objects
    :param j: QUA int
    :param unsafe: Passes the unsafe parameter to the QUA switch function. When set to true, fewer gaps will be created,
    but unexpected behavior will occur if j is not in `range(len(baking_list))`. Default is false
    """

    with qua.switch_(j, unsafe=unsafe):
        for i in range(len(baking_list)):
            with qua.case_(i):
                baking_list[i].run()


class BakingOperations:
    def __init__(self, b: Baking) -> None:
        super().__init__()
        self._baking = b

    def __getitem__(self, qe: str):
        return self._baking.get_op_name(qe)

    def length(self, qe):
        return self._baking.get_op_length(qe)



from qualang_tools.results.results import fetching_tool
from qualang_tools.results.results import progress_counter
from qualang_tools.results.results import wait_until_job_is_paused

from qualang_tools.results.data_handler import DataHandler, data_processors

__all__ = ["fetching_tool", "progress_counter", "wait_until_job_is_paused", "DataHandler", "data_processors"]



from scipy import optimize
from sklearn import preprocessing
import matplotlib.pyplot as plt
from typing import List, Union
import itertools
import json
import numpy as np
import warnings
from scipy.optimize import OptimizeWarning


class Fit:
    """
    This class takes care of the fitting to the measured data.
    It includes:
        - Fitting to: linear line
                      T1 experiment
                      Ramsey experiment
                      transmission resonator spectroscopy
                      reflection resonator spectroscopy
        - Printing the initial guess and fitting results
        - Plotting the data and the fitting function
        - Saving the data
    """

    # Remove optimize warnings
    warnings.simplefilter("ignore", RuntimeWarning)
    warnings.simplefilter("ignore", OptimizeWarning)

    @staticmethod
    def linear(
        x_data: Union[np.ndarray, List[float]],
        y_data: Union[np.ndarray, List[float]],
        guess=None,
        verbose=False,
        plot=False,
        save=False,
    ) -> dict:
        """
        Create a linear fit of the form

        .. math::
        f(x) = a * x + b

        for unknown parameters :
             a - The slope of the function
             b - The free parameter of the function

         :param x_data: The data on the x-axis
         :param y_data: The data on the y-axis
         :param dict guess: Dictionary containing the initial guess for the fitting parameters (guess=dict(a=20))
         :param verbose: if True prints the initial guess and fitting results
         :param plot: if True plots the data and the fitting function
         :param save: if not False saves the data into a json file
                      The id of the file is save='id'. The name of the json file is `id.json`
         :return: A dictionary of (fit_func, a, b)

        """

        # Normalizing the vectors
        xn = preprocessing.normalize([x_data], return_norm=True)
        yn = preprocessing.normalize([y_data], return_norm=True)
        x = xn[0][0]
        y = yn[0][0]
        x_normal = xn[1][0]
        y_normal = yn[1][0]

        # Finding an initial guess to the slope
        a0 = (y[-1] - y[0]) / (x[-1] - x[0])

        # Finding an initial guess to the free parameter
        b0 = y[0]

        # Check user guess
        if guess is not None:
            for key in guess.keys():
                if key == "a":
                    a0 = float(guess[key]) * x_normal / y_normal
                elif key == "b":
                    b0 = float(guess[key]) / y_normal
                else:
                    raise Exception(
                        f"The key '{key}' specified in 'guess' does not match a fitting parameters for this function."
                    )
        # Print the initial guess if verbose=True
        if verbose:
            print(f"Initial guess:\n" f" a = {a0 * y_normal / x_normal:.3f}, \n" f" b = {b0 * y_normal:.3f}")

        # Fitting function
        def func(x_var, c0, c1):
            return a0 * c0 * x_var + b0 * c1

        def fit_type(x_var, a):
            return func(x_var, a[0], a[1])

        popt, pcov = optimize.curve_fit(func, x, y, p0=[1, 1])
        perr = np.sqrt(np.diag(pcov))

        # Output the fitting function and its parameters
        out = {
            "fit_func": lambda x_var: fit_type(x_var / x_normal, popt) * y_normal,
            "a": [
                popt[0] * a0 * y_normal / x_normal,
                perr[0] * a0 * y_normal / x_normal,
            ],
            "b": [popt[1] * b0 * y_normal, perr[1] * b0 * y_normal],
        }
        # Print the fitting results if verbose=True
        if verbose:
            print(
                f"Fitting results:\n"
                f" a = {out['a'][0]:.3f} +/- {out['a'][1]:.3f}, \n"
                f" b = {out['b'][0]:.3f} +/- {out['b'][1]:.3f}"
            )
        # Plot the data and the fitting function if plot=True
        if plot:
            plt.plot(x_data, fit_type(x, popt) * y_normal)
            plt.plot(
                x_data,
                y_data,
                ".",
                label=f"a = {out['a'][0]:.1f} +/- {out['a'][1]:.1f} \n b = {out['b'][0]:.1f} +/- {out['b'][1]:.1f}",
            )
            plt.legend(loc="upper right")
        # Save the data in a json file named 'id.json' if save=id
        if save:
            fit_params = dict(itertools.islice(out.items(), 1, len(out)))
            fit_params["x_data"] = x_data.tolist()
            fit_params["y_data"] = y_data.tolist()
            fit_params["y_fit"] = (func(x, popt[0], popt[1]) * y_normal).tolist()
            json_object = json.dumps(fit_params)
            if save[-5:] == ".json":
                save = save[:-5]
            with open(f"{save}.json", "w") as outfile:
                outfile.write(json_object)

        return out

    @staticmethod
    def T1(
        x_data: Union[np.ndarray, List[float]],
        y_data: Union[np.ndarray, List[float]],
        guess=None,
        verbose=False,
        plot=False,
        save=False,
    ):
        """
        Create a fit to T1 experiment of the form

        .. math::
        f(x) = amp * np.exp(-x * (1/T1)) + final_offset

        for unknown parameters :
            T1 - The decay constant [ns]
            amp - The amplitude [a.u.]
            final_offset -  The offset visible for long waiting times [a.u.]

        :param x_data: The dephasing time [ns]
        :param y_data: Data containing the Ramsey signal
        :param dict guess: Dictionary containing the initial guess for the fitting parameters (guess=dict(T1=20))
        :param verbose: if True prints the initial guess and fitting results
        :param plot: if True plots the data and the fitting function
        :param save: if not False saves the data into a json file
                     The id of the file is save='id'. The name of the json file is `id.json`
        :return: A dictionary of (fit_func, T1, amp, final_offset)

        """
        # Normalizing the vectors
        xn = preprocessing.normalize([x_data], return_norm=True)
        yn = preprocessing.normalize([y_data], return_norm=True)
        x = xn[0][0]
        y = yn[0][0]
        x_normal = xn[1][0]
        y_normal = yn[1][0]

        # Finding a guess for the offsets
        final_offset = np.mean(y[np.min((int(len(y) * 0.9), len(y) - 3)) :])
        # Finding a guess for the decay
        guess_T1 = 1 / (np.abs(np.polyfit(x, np.log(np.abs(y - final_offset)), 1)[0])) / 2

        # Check user guess
        if guess is not None:
            for key in guess.keys():
                if key == "T1":
                    guess_T1 = float(guess[key]) / x_normal
                elif key == "amp":
                    pass
                elif key == "final_offset":
                    final_offset = float(guess[key]) / y_normal
                else:
                    raise Exception(
                        f"The key '{key}' specified in 'guess' does not match a fitting parameters for this function."
                    )
        # Print the initial guess if verbose=True
        if np.max(x_data) > 10:
            time_unit = "ns"
        else:
            time_unit = "s"
        if verbose:
            print(
                f"Initial guess:\n "
                f"T1 = {guess_T1 * x_normal:.3f} {time_unit}, \n "
                f"amp = {y[0] * y_normal:.3e} a.u., \n "
                f"final offset = {final_offset * y_normal:.3e} a.u."
            )

        # Fitting function
        def func(x_var, a0, a1, a2):
            return a1 * y[0] * np.exp(-x_var / (guess_T1 * a0)) + final_offset * a2

        def fit_type(x_var, a):
            return func(x_var, a[0], a[1], a[2])

        popt, pcov = optimize.curve_fit(
            func,
            x,
            y,
            p0=[1, 1, 1],
        )

        perr = np.sqrt(np.diag(pcov))

        # Output the fitting function and its parameters
        out = {
            "fit_func": lambda x_var: fit_type(x_var / x_normal, popt) * y_normal,
            "T1": [(guess_T1 * popt[0]) * x_normal, perr[0] * guess_T1 * x_normal],
            "amp": [popt[1] * y[0] * y_normal, perr[1] * y[0] * y_normal],
            "final_offset": [
                popt[2] * final_offset * y_normal,
                perr[2] * final_offset * y_normal,
            ],
        }
        # Print the fitting results if verbose=True
        if verbose:
            print(
                f"Fitting results:\n"
                f" T1 = {out['T1'][0]:.1f} +/- {out['T1'][1]:.1f} {time_unit}, \n"
                f" amp = {out['amp'][0]:.2e} +/- {out['amp'][1]:.1e} a.u., \n"
                f" final offset = {out['final_offset'][0]:.2e} +/- {out['final_offset'][1]:.1e} a.u."
            )
        # Plot the data and the fitting function if plot=True
        if plot:
            plt.plot(x_data, fit_type(x, popt) * y_normal)
            plt.plot(
                x_data,
                y_data,
                ".",
                label=f"T1 = {out['T1'][0]:.1f} +/- {out['T1'][1]:.1f} {time_unit}",
            )
            plt.legend(loc="upper right")
        # Save the data in a json file named 'id.json' if save=id
        if save:
            fit_params = dict(itertools.islice(out.items(), 1, len(out)))
            fit_params["x_data"] = x_data.tolist()
            fit_params["y_data"] = y_data.tolist()
            fit_params["y_fit"] = (fit_type(x, popt) * y_normal).tolist()
            json_object = json.dumps(fit_params)
            if save[-5:] == ".json":
                save = save[:-5]
            with open(f"{save}.json", "w") as outfile:
                outfile.write(json_object)

        return out

    @staticmethod
    def ramsey(
        x_data: Union[np.ndarray, List[float]],
        y_data: Union[np.ndarray, List[float]],
        guess=None,
        verbose=False,
        plot=False,
        save=False,
    ):
        """
        Create a fit to Ramsey experiment of the form

        .. math::
        f(x) = final_offset * (1 - np.exp(-x * (1/T2))) + amp / 2 * (
            np.exp(-x * (1/T2))
            * (initial_offset * 2 + np.cos(2 * np.pi * f * x + phase))
            )

        for unknown parameters :
            f - The detuning frequency [GHz]
            phase - The phase [rad]
            T2 - The decay constant [ns]
            amp - The amplitude [a.u.]
            final_offset -  The offset visible for long dephasing times [a.u.]
            initial_offset - The offset visible for short dephasing times

        :param x_data: The dephasing time [ns]
        :param y_data: Data containing the Ramsey signal
        :param dict guess: Dictionary containing the initial guess for the fitting parameters (guess=dict(T2=20))
        :param verbose: if True prints the initial guess and fitting results
        :param plot: if True plots the data and the fitting function
        :param save: if not False saves the data into a json file
                     The id of the file is save='id'. The name of the json file is `id.json`
          :return: A dictionary of (fit_func, f, phase, tau, amp, uncertainty_population, initial_offset)

        """

        # Normalizing the vectors
        xn = preprocessing.normalize([x_data], return_norm=True)
        yn = preprocessing.normalize([y_data], return_norm=True)
        x = xn[0][0]
        y = yn[0][0]
        x_normal = xn[1][0]
        y_normal = yn[1][0]

        # Compute the FFT for guessing the frequency
        fft = np.fft.fft(y)
        f = np.fft.fftfreq(len(x))
        # Take the positive part only
        fft = fft[1 : len(f) // 2]
        f = f[1 : len(f) // 2]
        # Remove the DC peak if there is one
        if (np.abs(fft)[1:] - np.abs(fft)[:-1] > 0).any():
            first_read_data_ind = np.where(np.abs(fft)[1:] - np.abs(fft)[:-1] > 0)[0][0]  # away from the DC peak
            fft = fft[first_read_data_ind:]
            f = f[first_read_data_ind:]

        # Finding a guess for the frequency
        out_freq = f[np.argmax(np.abs(fft))]
        guess_freq = out_freq / (x[1] - x[0])

        # The period is 1 / guess_freq --> number of oscillations --> peaks decay to get guess_T2
        period = int(np.ceil(1 / out_freq))
        peaks = (
            np.array([np.std(y[i * period : (i + 1) * period]) for i in range(round(len(y) / period))]) * np.sqrt(2) * 2
        )

        # Finding a guess for the decay (slope of log(peaks))
        if len(peaks) > 1:
            guess_T2 = -1 / ((np.log(peaks)[-1] - np.log(peaks)[0]) / (period * (len(peaks) - 1))) * (x[1] - x[0])
        else:
            guess_T2 = 100 / x_normal

        # Finding a guess for the offsets
        initial_offset = np.mean(y[:period])
        final_offset = np.mean(y[-period:])

        # Finding a guess for the phase
        guess_phase = np.angle(fft[np.argmax(np.abs(fft))]) - guess_freq * 2 * np.pi * x[0]

        # Check user guess
        if guess is not None:
            for key in guess.keys():
                if key == "f":
                    guess_freq = float(guess[key]) * x_normal
                elif key == "phase":
                    guess_phase = float(guess[key])
                elif key == "T2":
                    guess_T2 = float(guess[key]) * x_normal
                elif key == "amp":
                    peaks[0] = float(guess[key]) / y_normal
                elif key == "initial_offset":
                    initial_offset = float(guess[key]) / y_normal
                elif key == "final_offset":
                    final_offset = float(guess[key]) / y_normal
                else:
                    raise Exception(
                        f"The key '{key}' specified in 'guess' does not match a fitting parameters for this function."
                    )

        # Print the initial guess if verbose=True
        if verbose:
            print(
                f"Initial guess:\n"
                f" f = {guess_freq / x_normal:.3f}, \n"
                f" phase = {guess_phase:.3f}, \n"
                f" T2 = {guess_T2 * x_normal:.3f}, \n"
                f" amp = {peaks[0] * y_normal:.3f}, \n"
                f" initial offset = {initial_offset * y_normal:.3f}, \n"
                f" final_offset = {final_offset * y_normal:.3f}"
            )

        # Fitting function
        def func(x_var, a0, a1, a2, a3, a4, a5):
            return final_offset * a4 * (1 - np.exp(-x_var / (guess_T2 * a1))) + peaks[0] / 2 * a2 * (
                np.exp(-x_var / (guess_T2 * a1))
                * (a5 * initial_offset / peaks[0] * 2 + np.cos(2 * np.pi * a0 * guess_freq * x + a3))
            )

        def fit_type(x_var, a):
            return func(x_var, a[0], a[1], a[2], a[3], a[4], a[5])

        popt, pcov = optimize.curve_fit(
            func,
            x,
            y,
            p0=[1, 1, 1, guess_phase, 1, 1],
        )

        perr = np.sqrt(np.diag(pcov))

        # Output the fitting function and its parameters
        out = {
            "fit_func": lambda x_var: fit_type(x_var / x_normal, popt) * y_normal,
            "f": [popt[0] * guess_freq / x_normal, perr[0] * guess_freq / x_normal],
            "phase": [popt[3] % (2 * np.pi), perr[3] % (2 * np.pi)],
            "T2": [(guess_T2 * popt[1]) * x_normal, perr[1] * guess_T2 * x_normal],
            "amp": [peaks[0] * popt[2] * y_normal, perr[2] * peaks[0] * y_normal],
            "initial_offset": [
                popt[5] * initial_offset * y_normal,
                perr[5] * initial_offset * y_normal,
            ],
            "final_offset": [
                final_offset * popt[4] * y_normal,
                perr[4] * final_offset * y_normal,
            ],
        }
        # Print the fitting results if verbose=True
        if verbose:
            print(
                f"Fitting results:\n"
                f" f = {out['f'][0] * 1000:.3f} +/- {out['f'][1] * 1000:.3f} MHz, \n"
                f" phase = {out['phase'][0]:.3f} +/- {out['phase'][1]:.3f} rad, \n"
                f" T2 = {out['T2'][0]:.2f} +/- {out['T2'][1]:.3f} ns, \n"
                f" amp = {out['amp'][0]:.2f} +/- {out['amp'][1]:.3f} a.u., \n"
                f" initial offset = {out['initial_offset'][0]:.2f} +/- {out['initial_offset'][1]:.3f}, \n"
                f" final_offset = {out['final_offset'][0]:.2f} +/- {out['final_offset'][1]:.3f} a.u."
            )
        # Plot the data and the fitting function if plot=True
        if plot:
            plt.plot(x_data, fit_type(x, popt) * y_normal)
            plt.plot(
                x_data,
                y_data,
                ".",
                label=f"T2 = {out['T2'][0]:.1f} +/- {out['T2'][1]:.1f}ns \n f = {out['f'][0] * 1000:.3f} +/- {out['f'][1] * 1000:.3f} MHz",
            )
            plt.legend(loc="upper right")
        # Save the data in a json file named 'id.json' if save=id
        if save:
            fit_params = dict(itertools.islice(out.items(), 1, len(out)))
            fit_params["x_data"] = x_data.tolist()
            fit_params["y_data"] = y_data.tolist()
            fit_params["y_fit"] = (fit_type(x, popt) * y_normal).tolist()
            json_object = json.dumps(fit_params)
            if save[-5:] == ".json":
                save = save[:-5]
            with open(f"{save}.json", "w") as outfile:
                outfile.write(json_object)
        return out

    @staticmethod
    def transmission_resonator_spectroscopy(
        x_data: Union[np.ndarray, List[float]],
        y_data: Union[np.ndarray, List[float]],
        guess=None,
        verbose=False,
        plot=False,
        save=False,
    ):
        """
        Create a fit to transmission resonator spectroscopy of the form

        .. math::
        ((kc/k) / (
            1 + (4 * ((x - f) ** 2) / (k ** 2)))) + offset

        for unknown parameters:
            f - The frequency at the peak
            kc - The strength with which the field of the resonator couples to the transmission line
            ki - A parameter that indicates the internal coherence properties of the resonator
            k - The FWHM of the fitted function.  k = ki + kc
            offset - The offset

        :param x_data:  The frequency in Hz
        :param y_data: The transition probability (I^2+Q^2)
        :param dict guess: Dictionary containing the initial guess for the fitting parameters (guess=dict(f=20e6))
        :param verbose: if True prints the initial guess and fitting results
        :param plot: if True plots the data and the fitting function
        :param save: if not False saves the data into a json file
                     The id of the file is save='id'. The name of the json file is `id.json`
             :return: A dictionary of (fit_func, f, kc, k, ki, offset)

        """

        xn = preprocessing.normalize([x_data], return_norm=True)
        yn = preprocessing.normalize([y_data], return_norm=True)
        x = xn[0][0]
        y = yn[0][0]
        x_normal = xn[1][0]
        y_normal = yn[1][0]

        # Finding a guess for the max
        peak = max(y)
        arg_max = y.argmax()

        # Finding an initial guess for the FWHM
        if arg_max > len(y) / 2:
            y_FWHM = (peak + np.mean(y[0:10])) / 2
        else:
            y_FWHM = (peak + np.mean(y[-10:-1])) / 2

        # Finding a guess for the width
        width0_arg_right = (np.abs(y_FWHM - y[arg_max + 1 : len(y)])).argmin() + arg_max
        width0_arg_left = (np.abs(y_FWHM - y[0:arg_max])).argmin()
        width0 = x[width0_arg_right] - x[width0_arg_left]

        # Finding the frequency at the min
        f0 = x[arg_max]

        # Finding a guess to offset
        v0 = np.mean(y[0 : int(width0_arg_left - width0 / 2)])

        # Check user guess
        if guess is not None:
            for key in guess.keys():
                if key == "f":
                    f0 = float(guess[key]) / x_normal
                elif key == "k":
                    width0 = float(guess[key]) / x_normal
                elif key == "offset":
                    v0 = float(guess[key]) / y_normal
                else:
                    raise Exception(
                        f"The key '{key}' specified in 'guess' does not match a fitting parameters for this function."
                    )
        # Print the initial guess if verbose=True
        if verbose:
            print(
                f"Initial guess:\n "
                f" f = {f0 * x_normal}, \n "
                f" kc = {(peak - v0) * (width0 * x_normal) * y_normal}, \n "
                f" k = {width0 * x_normal}, \n "
                f" offset = {v0 * y_normal}"
            )

        # Fitting function
        def func(x_var, a0, a1, a2, a3):
            return (((peak - v0) * a0) / (1 + (4 * ((x_var - (f0 * a2)) ** 2) / ((width0 * a1) ** 2)))) + (v0 * a3)

        def fit_type(x_var, a):
            return func(x_var, a[0], a[1], a[2], a[3])

        popt, pcov = optimize.curve_fit(func, x, y, p0=[1, 1, 1, 1])
        perr = np.sqrt(np.diag(pcov))

        # Output the fitting function and its parameters
        out = {
            "fit_func": lambda x_var: fit_type(x_var / x_normal, popt) * y_normal,
            "f": [f0 * popt[2] * x_normal, f0 * perr[2] * x_normal],
            "kc": [
                (peak - v0) * popt[0] * (width0 * popt[1] * x_normal) * y_normal,
                (peak - v0) * perr[0] * (width0 * perr[1] * x_normal) * y_normal,
            ],
            "ki": [
                (popt[1] * width0 * x_normal) - ((peak - v0) * popt[0] * (width0 * popt[1] * x_normal) * y_normal),
                (perr[1] * width0 * x_normal) - ((peak - v0) * perr[0] * (width0 * perr[1] * x_normal) * y_normal),
            ],
            "k": [popt[1] * width0 * x_normal, perr[1] * width0 * x_normal],
            "offset": [v0 * popt[3] * y_normal, v0 * perr[3] * y_normal],
        }
        # Check freq unit
        if np.max(np.abs(x_data)) < 18:
            freq_unit = "GHz"
        elif np.max(np.abs(x_data)) < 18000:
            freq_unit = "MHz"
        else:
            freq_unit = "Hz"
        # Print the fitting results if verbose=True
        if verbose:
            print(
                f"Fit results:\n"
                f"f = {out['f'][0]:.3f} +/- {out['f'][1]:.3f} {freq_unit}, \n"
                f"kc = {out['kc'][0]:.3f} +/- {out['kc'][1]:.3f} {freq_unit}, \n"
                f"ki = {out['ki'][0]:.3f} +/- {out['ki'][1]:.3f} {freq_unit}, \n"
                f"k = {out['k'][0]:.3f} +/- {out['k'][1]:.3f} {freq_unit}, \n"
                f"offset = {out['offset'][0]:.3f} +/- {out['offset'][1]:.3f} a.u., \n"
            )
        # Plot the data and the fitting function if plot=True
        if plot:
            plt.plot(x_data, fit_type(x, popt) * y_normal)
            plt.plot(
                x_data,
                y_data,
                ".",
                label=f"f = {out['f'][0]:.1f} +/- {out['f'][1]:.1f} {freq_unit}",
            )
            plt.xlabel(f"Frequency {freq_unit}")
            plt.legend(loc="upper right")
        # Save the data in a json file named 'id.json' if save=id
        if save:
            fit_params = dict(itertools.islice(out.items(), 1, len(out)))
            fit_params["x_data"] = x_data.tolist()
            fit_params["y_data"] = y_data.tolist()
            fit_params["y_fit"] = (fit_type(x, popt) * y_normal).tolist()
            json_object = json.dumps(fit_params)
            if save[-5:] == ".json":
                save = save[:-5]
            with open(f"{save}.json", "w") as outfile:
                outfile.write(json_object)

        return out

    @staticmethod
    def reflection_resonator_spectroscopy(
        x_data: Union[np.ndarray, List[float]],
        y_data: Union[np.ndarray, List[float]],
        guess=None,
        verbose=False,
        plot=False,
        save=False,
    ):
        """
        Create a fit to reflection resonator spectroscopy of the form

        .. math::
        (offset)-((kc/k) / (
            1 + (4 * ((x - f) ** 2) / (k ** 2)))) + slope * x
        for unknown parameters:
            f - The frequency at the peak
            kc - The strength with which the field of the resonator couples to the transmission line
            ki - A parameter that indicates the internal coherence properties of the resonator
            k - The FWHM of the fitted function.  k = ki + kc
            offset - The offset
            slope - The slope of the function. This is added after experimental considerations.

        :param x_data:  The frequency in Hz
        :param y_data: The transition probability (I^2+Q^2)
        :param dict guess: Dictionary containing the initial guess for the fitting parameters (guess=dict(f=20e6))
        :param verbose: if True prints the initial guess and fitting results
        :param plot: if True plots the data and the fitting function
        :param save: if not False saves the data into a json file
                     The id of the file is save='id'. The name of the json file is `id.json`
          :return: A dictionary of (fit_func, f, kc, k, ki, offset)

        """

        # Normalizing the vectors
        xn = preprocessing.normalize([x_data], return_norm=True)
        yn = preprocessing.normalize([y_data], return_norm=True)
        x = xn[0][0]
        y = yn[0][0]
        x_normal = xn[1][0]
        y_normal = yn[1][0]

        # Finding a guess for the min
        peak = min(y)
        arg_min = y.argmin()

        # Finding the frequency at the min
        f0 = x[arg_min]

        # Finding an initial guess for the FWHM
        if arg_min > len(y) / 2:
            y_FWHM = (peak + np.mean(y[0:10])) / 2
        else:
            y_FWHM = (peak + np.mean(y[-10:-1])) / 2

        # Finding a guess to the width
        width0_arg_right = (np.abs(y_FWHM - y[arg_min + 1 : len(y)])).argmin() + arg_min
        width0_arg_left = (np.abs(y_FWHM - y[0:arg_min])).argmin()
        width0 = x[width0_arg_right] - x[width0_arg_left]
        width0 = width0
        # Finding guess to offset
        v0 = (np.mean(y[-10:-1]) + np.mean(y[0:10])) / 2

        # Finding a guess to the slope
        m = (np.mean(y[int(width0_arg_right + width0) : -1]) - np.mean(y[0 : int(width0_arg_left - width0)])) / (
            np.mean(x[int(width0_arg_right + width0) : -1]) - np.mean(x[0 : int(width0_arg_left - width0)])
        )

        # Check user guess
        if guess is not None:
            for key in guess.keys():
                if key == "f":
                    f0 = float(guess[key]) / x_normal
                elif key == "k":
                    width0 = float(guess[key]) / x_normal
                elif key == "offset":
                    v0 = float(guess[key]) / y_normal
                elif key == "slope":
                    m = float(guess[key]) / y_normal * x_normal
                else:
                    raise Exception(
                        f"The key '{key}' specified in 'guess' does not match a fitting parameters for this function."
                    )

        # Print the initial guess if verbose=True
        if verbose:
            print(
                f"Initial guess:\n "
                f" f = {f0 * x_normal}, \n "
                f" kc = {(v0 - peak) * (width0 * x_normal) * y_normal}, \n "
                f" k = {width0 * x_normal}, \n "
                f" offset = {v0 * y_normal}, \n "
                f" slope = {m * y_normal / x_normal} "
            )

        def func(x_var, a0, a1, a2, a3, a4):
            return (
                ((v0 - peak) * a3)
                - (((v0 - peak) * a0) / (1 + (4 * ((x_var - (f0 * a2)) ** 2) / ((width0 * a1) ** 2))))
                + m * a4 * x_var
            )

        def fit_type(x_var, a):
            return func(x_var, a[0], a[1], a[2], a[3], a[4])

        popt, pcov = optimize.curve_fit(func, x, y, p0=[1, 1, 1, 1, 1])
        perr = np.sqrt(np.diag(pcov))

        # Output the fitting function and its parameters
        out = {
            "fit_func": lambda x_var: fit_type(x_var / x_normal, popt) * y_normal,
            "f": [f0 * popt[2] * x_normal, f0 * perr[2] * x_normal],
            "kc": [
                (v0 - peak) * popt[0] * (width0 * popt[1] * x_normal) * y_normal,
                (v0 - peak) * perr[0] * (width0 * perr[1] * x_normal) * y_normal,
            ],
            "ki": [
                (popt[1] * width0 * x_normal) - ((v0 - peak) * popt[0] * (width0 * popt[1] * x_normal) * y_normal),
                (perr[1] * width0 * x_normal) - ((v0 - peak) * perr[0] * (width0 * perr[1] * x_normal) * y_normal),
            ],
            "k": [popt[1] * width0 * x_normal, perr[1] * width0 * x_normal],
            "offset": [
                (v0 - peak) * popt[3] * y_normal,
                (v0 - peak) * perr[3] * y_normal,
            ],
            "slope": [
                m * popt[4] * y_normal / x_normal,
                m * perr[4] * y_normal / x_normal,
            ],
        }
        # Check freq unit
        if np.max(np.abs(x_data)) < 18:
            freq_unit = "GHz"
        elif np.max(np.abs(x_data)) < 18000:
            freq_unit = "MHz"
        else:
            freq_unit = "Hz"
        # Print the fitting results if verbose=True
        if verbose:
            print(
                f"Fit results:\n"
                f"f = {out['f'][0]:.3f} +/- {out['f'][1]:.3f} {freq_unit}, \n"
                f"kc = {out['kc'][0]:.3f} +/- {out['kc'][1]:.3f} {freq_unit}, \n"
                f"ki = {out['ki'][0]:.3f} +/- {out['ki'][1]:.3f} {freq_unit}, \n"
                f"k = {out['k'][0]:.3f} +/- {out['k'][1]:.3f} {freq_unit}, \n"
                f"offset = {out['offset'][0]:.3f} +/- {out['offset'][1]:.3f} a.u., \n"
                f"slope = {out['slope'][0]:.3f} +/- {out['slope'][1]:.3f} 1/{freq_unit}\n"
            )
        # Plot the data and the fitting function if plot=True
        if plot:
            plt.plot(x_data, fit_type(x, popt) * y_normal)
            plt.plot(
                x_data,
                y_data,
                ".",
                label=f"f = {out['f'][0]:.1f} +/- {out['f'][1]:.1f} {freq_unit}",
            )
            plt.legend(loc="upper right")
        # Save the data in a json file named 'id.json' if save=id
        if save:
            fit_params = dict(itertools.islice(out.items(), 1, len(out)))
            fit_params["x_data"] = x_data.tolist()
            fit_params["y_data"] = y_data.tolist()
            fit_params["y_fit"] = (fit_type(x, popt) * y_normal).tolist()
            json_object = json.dumps(fit_params)
            if save[-5:] == ".json":
                save = save[:-5]
            with open(f"{save}.json", "w") as outfile:
                outfile.write(json_object)

        return out

    @staticmethod
    def rabi(
        x_data: Union[np.ndarray, List[float]],
        y_data: Union[np.ndarray, List[float]],
        guess=None,
        verbose=False,
        plot=False,
        save=False,
    ):
        """
        Create a fit to Rabi experiment of the form

        .. math::
        f(x) = amp * (np.sin(0.5 * (2 * np.pi * f) * x_var + phase))**2 * np.exp(-x_var / T) + offset

        for unknown parameters :
            f - The detuning frequency [GHz]
            phase - The phase [rad]
            T - The decay constant [ns]
            amp - The amplitude [a.u.]
            offset -  The offset visible for long dephasing times [a.u.]

        :param x_data: The dephasing time [ns]
        :param y_data: Data containing the Ramsey signal
        :param dict guess: Dictionary containing the initial guess for the fitting parameters (guess=dict(T2=20))
        :param verbose: if True prints the initial guess and fitting results
        :param plot: if True plots the data and the fitting function
        :param save: if not False saves the data into a json file
                     The id of the file is save='id'. The name of the json file is `data_fit_id.json`
          :return: A dictionary of (fit_func, f, phase, tau, amp, uncertainty_population, initial_offset)

        """

        # Normalizing the vectors
        xn = preprocessing.normalize([x_data], return_norm=True)
        yn = preprocessing.normalize([y_data], return_norm=True)
        x = xn[0][0]
        y = yn[0][0]
        x_normal = xn[1][0]
        y_normal = yn[1][0]

        # Compute the FFT for guessing the frequency
        fft = np.fft.fft(y)
        f = np.fft.fftfreq(len(x))
        # Take the positive part only
        fft = fft[1 : len(f) // 2]
        f = f[1 : len(f) // 2]

        # Finding a guess for the frequency
        out_freq = f[np.argmax(np.abs(fft))]
        guess_freq = out_freq / (x[1] - x[0])
        # If not enough oscillations
        if np.argmax(np.abs(fft)) < 1:
            # If at least 1 oscillation
            if np.max(np.diff(np.where(y < min(y) + 0.1 * np.abs(max(y) - min(y))))) > 1:
                guess_freq = 1 / (np.max(np.diff(np.where(y < min(y) + 0.1 * np.abs(max(y) - min(y))))) * (x[1] - x[0]))
            # If less than 1 oscillation
            else:
                guess_freq = 1 / (
                    (np.argmin(y[len(y) // 2 :]) + len(y) // 2 - np.argmin(y[: len(y) // 2])) * (x[1] - x[0])
                )
        # The period is 1 / guess_freq --> number of oscillations --> peaks decay to get guess_T
        period = int(np.ceil(1 / out_freq))
        peaks = (
            np.array([np.std(y[i * period : (i + 1) * period]) for i in range(round(len(y) / period))]) * np.sqrt(2) * 2
        )

        # Finding a guess for the decay (slope of log(peaks))
        if len(peaks) > 1:
            guess_T = -1 / ((np.log(peaks)[-1] - np.log(peaks)[0]) / (period * (len(peaks) - 1))) * (x[1] - x[0])
        else:
            guess_T = 100 / x_normal

        # Finding a guess for the offset
        offset = np.mean(y[-period:])

        # Finding a guess for the offset
        guess_amp = np.abs(np.max(y) - np.min(y))

        # Finding a guess for the phase
        guess_phase = np.angle(fft[np.argmax(np.abs(fft))]) - guess_freq * 2 * np.pi * x[0]

        # Check user guess
        if guess is not None:
            for key in guess.keys():
                if key == "f":
                    guess_freq = float(guess[key]) * x_normal
                elif key == "phase":
                    guess_phase = float(guess[key])
                elif key == "T":
                    guess_T = float(guess[key]) * x_normal
                elif key == "amp":
                    guess_amp = float(guess[key]) / y_normal
                elif key == "offset":
                    offset = float(guess[key]) / y_normal
                else:
                    raise Exception(
                        f"The key '{key}' specified in 'guess' does not match a fitting parameters for this function."
                    )

        # Print the initial guess if verbose=True
        if verbose:
            print(
                f"Initial guess:\n"
                f" f = {guess_freq / x_normal:.3f}, \n"
                f" phase = {guess_phase:.3f}, \n"
                f" T = {guess_T * x_normal:.3f}, \n"
                f" amplitude = {guess_amp * y_normal:.1e}, \n"
                f" offset = {offset * y_normal:.1e}"
            )

        # Fitting function
        def func(x_var, a0, a1, a2, a3, a4):
            return (guess_amp * a0) * (np.sin(0.5 * (2 * np.pi * guess_freq * a1) * x_var + a3)) ** 2 * np.exp(
                -x_var / np.abs(guess_T * a2)
            ) + offset * a4

        def fit_type(x_var, a):
            return func(x_var, a[0], a[1], a[2], a[3], a[4])

        popt, pcov = optimize.curve_fit(
            func,
            x,
            y,
            p0=[1, 1, 1, guess_phase, 1],
        )
        perr = np.sqrt(np.diag(pcov))

        # Output the fitting function and its parameters
        out = {
            "fit_func": lambda x_var: fit_type(x_var / x_normal, popt) * y_normal,
            "f": [popt[1] * guess_freq / x_normal, perr[1] * guess_freq / x_normal],
            "phase": [popt[3] % (2 * np.pi), perr[3] % (2 * np.pi)],
            "T": [(guess_T * abs(popt[2])) * x_normal, perr[2] * guess_T * x_normal],
            "amp": [popt[0] * guess_amp * y_normal, perr[0] * guess_amp * y_normal],
            "offset": [
                offset * popt[4] * y_normal,
                perr[4] * offset * y_normal,
            ],
        }
        # FInd if power rabi or time rabi
        if 0.01 < np.abs(np.max(x_data)) < 1:
            pi_param = 0.5 / out["f"][0]
            pi_label = f"x180 amplitude = {pi_param:.3f} V"
        else:
            pi_param = 0.5 / out["f"][0]
            if pi_param > 1:
                pi_label = f"x180 length = {pi_param:.1f} ns"
            else:
                pi_label = f"x180 length = {pi_param:.3e} s"
        # Print the fitting results if verbose=True
        if verbose:
            print(
                f"Fitting results:\n"
                f" f = {out['f'][0] * 1000:.3f} +/- {out['f'][1] * 1000:.3f} MHz, \n"
                f" phase = {out['phase'][0]:.3f} +/- {out['phase'][1]:.3f} rad, \n"
                f" T = {out['T'][0]:.2f} +/- {out['T'][1]:.3f} ns, \n"
                f" amplitude = {out['amp'][0]:.1e} +/- {out['amp'][1]:.1e} a.u., \n"
                f" offset = {out['offset'][0]:.1e} +/- {out['offset'][1]:.1e} a.u."
            )
        # Plot the data and the fitting function if plot=True
        if plot:
            plt.plot(x_data, fit_type(x, popt) * y_normal)
            plt.plot(
                x_data,
                y_data,
                ".",
                label=pi_label,
            )
            plt.legend(loc="upper right")
        # Save the data in a json file named 'data_fit_id.json' if save=id
        if save:
            fit_params = dict(itertools.islice(out.items(), 1, len(out)))
            fit_params["x_data"] = x_data.tolist()
            fit_params["y_data"] = y_data.tolist()
            fit_params["y_fit"] = (fit_type(x, popt) * y_normal).tolist()
            json_object = json.dumps(fit_params)
            with open(f"data_fit_{save}.json", "w") as outfile:
                outfile.write(json_object)
        return out

    @staticmethod
    def resonator_frequency_vs_flux(
        frequency: Union[np.ndarray, List[float]],
        flux: Union[np.ndarray, List[float]],
        data: np.ndarray,
        guess=None,
        verbose=False,
        plot=False,
        save=False,
    ):
        """
        Create a fit for the resonator frequency versus flux 2D map.
        Each flux points (line-cuts) is fitted with the reflection_resonator_spectroscopy() function to extract the
        resonance frequency of the resonator and the evolution of this frequency versu flux if fitted using

        .. math::
        f(x) = amp * (np.sin((2 * np.pi * f) * frequency + phase))**2 + offset

        for unknown parameters :
            f - The resonator vs flux oscillating frequency [Hz or GHz / V]
            phase - The phase [rad]
            amp - The oscillation amplitude [Hz or GHz]
            offset -  The resonator frequency offset [Hz or GHz]

        :param frequency: The readout frequency [Hz or GHz]
        :param flux: The flux biases [V]
        :param data: Data containing the frequency vs flux map. Must be a 2D array with data.shape=(len(flux), len(frequency)).
        :param dict guess: Dictionary containing the initial guess for the fitting parameters (guess=dict(T2=20))
        :param verbose: if True prints the initial guess and fitting results
        :param plot: if True plots the data and the fitting function
        :param save: if not False saves the data into a json file
                     The id of the file is save='id'. The name of the json file is `data_fit_id.json`
          :return: A dictionary of (fit_func, f, phase, amp, offset)

        """
        freq = [np.nan for _ in range(len(flux))]
        # Extract 1D curve f_res vs flux
        map2 = np.zeros((len(flux), len(frequency)))
        for i in range(len(flux)):
            try:
                f = Fit.reflection_resonator_spectroscopy(frequency, data[i])
                freq[i] = f["f"][0]
            except (Exception,):
                if i > 0:
                    freq[i] = freq[i - 1]
                else:
                    freq[i] = np.mean(frequency)
            # Get fitted curve on 2D map
            map2[i][np.argmin(np.abs(frequency - freq[i]))] = -np.max(data)
        # Normalizing the vectors
        xn = preprocessing.normalize([flux], return_norm=True)
        yn = preprocessing.normalize([freq], return_norm=True)
        x = xn[0][0]
        y = yn[0][0]
        x_normal = xn[1][0]
        y_normal = yn[1][0]

        # Compute the FFT for guessing the frequency
        fft = np.fft.fft(y)
        f = np.fft.fftfreq(len(x))
        # Take the positive part only
        fft = fft[1 : len(f) // 2]
        f = f[1 : len(f) // 2]
        # Finding a guess for the frequency
        out_freq = f[np.argmax(np.abs(fft))]
        guess_freq = out_freq / (x[1] - x[0])
        if np.argmax(np.abs(fft)) < 1:
            guess_freq = 1 / (x[-1] - x[0])

        # Finding a guess for the offset
        offset = np.mean(y)

        # Finding a guess for the phase
        guess_phase = np.angle(fft[np.argmax(np.abs(fft))]) - guess_freq * 2 * np.pi * x[0]

        guess_amp = np.max(y) - np.min(y)
        # Check user guess
        if guess is not None:
            for key in guess.keys():
                if key == "f":
                    guess_freq = float(guess[key]) * x_normal
                elif key == "phase":
                    guess_phase = float(guess[key])
                elif key == "amp":
                    guess_amp = float(guess[key]) / y_normal
                elif key == "offset":
                    offset = float(guess[key]) / y_normal
                else:
                    raise Exception(
                        f"The key '{key}' specified in 'guess' does not match a fitting parameters for this function."
                    )

        # Print the initial guess if verbose=True
        if verbose:
            print(
                f"Initial guess:\n"
                f" f = {guess_freq / x_normal:.3f}, \n"
                f" phase = {guess_phase:.3f}, \n"
                f" amplitude = {guess_amp * y_normal:.3f}, \n"
                f" offset = {offset * y_normal:.3f}"
            )

        # Fitting function
        def func(x_var, a0, a1, a2, a3):
            return (guess_amp * a0) * np.sin((2 * np.pi * guess_freq * a1) * x_var + a2) + offset * a3

        def fit_type(x_var, a):
            return func(x_var, a[0], a[1], a[2], a[3])

        popt, pcov = optimize.curve_fit(
            func,
            x,
            y,
            p0=[1, 1, guess_phase, 1],
        )
        perr = np.sqrt(np.diag(pcov))
        # Check freq unit
        if np.max(np.abs(frequency)) < 18:
            freq_unit = "GHz"
        elif np.max(np.abs(frequency)) < 18000:
            freq_unit = "MHz"
        else:
            freq_unit = "Hz"
        # Output the fitting function and its parameters
        out = {
            "fit_func": lambda x_var: fit_type(x_var / x_normal, popt) * y_normal,
            "f": [popt[1] * guess_freq / x_normal, perr[1] * guess_freq / x_normal],
            "phase": [popt[2] % (2 * np.pi), perr[2] % (2 * np.pi)],
            "amp": [popt[0] * guess_amp * y_normal, perr[0] * guess_amp * y_normal],
            "offset": [
                offset * popt[3] * y_normal,
                perr[3] * offset * y_normal,
            ],
        }
        # Print the fitting results if verbose=True
        if verbose:
            print(
                f"Fitting results:\n"
                f" f = {out['f'][0]:.3f} +/- {out['f'][1]:.3f} 1/V, \n"
                f" phase = {out['phase'][0]:.3f} +/- {out['phase'][1]:.3f} rad, \n"
                f" amplitude = {out['amp'][0]:.2f} +/- {out['amp'][1]:.3f} {freq_unit}, \n"
                f" offset = {out['offset'][0]:.2f} +/- {out['offset'][1]:.3f} {freq_unit}"
            )
        # Plot the data and the fitting function if plot=True
        if plot:
            plt.figure()
            plt.subplot(211)
            plt.pcolor(frequency, flux, map2 + data)
            plt.ylabel("Flux bias [V]")
            plt.xlabel(f"Readout frequency [{freq_unit}]")
            plt.title("Readout resonator spec vs flux")
            plt.subplot(212)
            plt.plot(flux, freq, "b.")
            plt.plot(flux, fit_type(x, popt) * y_normal)
            plt.tight_layout()
            if np.max(np.abs(flux)) < 1:
                plt.xlabel("Flux bias [V]")
            else:
                plt.xlabel("Flux bias [mV]")
            plt.ylabel(f"Resonator frequency [{freq_unit}]")
            plt.tight_layout()
        # Save the data in a json file named 'data_fit_id.json' if save=id
        if save:
            fit_params = dict(itertools.islice(out.items(), 1, len(out)))
            fit_params["x_data"] = flux.tolist()
            fit_params["y_data"] = freq
            fit_params["y_fit"] = (fit_type(x, popt) * y_normal).tolist()
            json_object = json.dumps(fit_params)
            with open(f"data_fit_{save}.json", "w") as outfile:
                outfile.write(json_object)
        return out


class Read:
    """
    This class takes care of reading the saved data.
    """

    @staticmethod
    def read_saved_params(file_id: str, verbose=False) -> dict:
        """
        Read the saved json file and print the saved params if print_params=True
        :param file_id: is the name of the json file as been given in the Fit class
        :param verbose: The parameters that were saved
        :return: Dictionary containing the saved data
        """
        if file_id[-5:] == ".json":
            file_id = file_id[:-5]
        f = open(f"{file_id}.json")
        data = json.load(f)
        if verbose:
            for key, value in data.items():
                print("{} = {}".format(key, value))
        return data



from .channel_specs import (
    ChannelSpecMwFemSingle,
    ChannelSpecLfFemSingle,
    ChannelSpecLfFemBaseband,
    ChannelSpecLfFemBasebandAndOctave,
    ChannelSpecOpxPlusSingle,
    ChannelSpecOpxPlusBaseband,
    ChannelSpecOpxPlusBasebandAndOctave,
    ChannelSpecOctave,
)

mw_fem_spec = ChannelSpecMwFemSingle
lf_fem_spec = ChannelSpecLfFemSingle
lf_fem_iq_spec = ChannelSpecLfFemBaseband
lf_fem_iq_octave_spec = ChannelSpecLfFemBasebandAndOctave
opx_spec = ChannelSpecOpxPlusSingle
opx_iq_spec = ChannelSpecOpxPlusBaseband
opx_iq_octave_spec = ChannelSpecOpxPlusBasebandAndOctave
octave_spec = ChannelSpecOctave
